import importlib
import orjson

from . import models


COUNTRIES_BY_NAME = {}
COUNTRIES_BY_ISO = {}
CITIES_BY_GEONAME_ID = {}
CITIES_BY_COUNTRY = {}


with importlib.resources.path("archon", "geodata") as geodata:
    with importlib.resources.as_file(geodata / "countries.json") as countries:
        COUNTRIES_BY_NAME = {
            d["country"]: models.Country(**d)
            for d in orjson.loads(countries.read_bytes())
        }
        COUNTRIES_BY_ISO = {c.iso: c for c in COUNTRIES_BY_NAME.values()}

    with importlib.resources.as_file(geodata / "cities.json") as cities:
        CITIES_BY_COUNTRY = {
            country: {name: models.City(**data) for name, data in cities.items()}
            for country, cities in orjson.loads(cities.read_bytes()).items()
        }
        CITIES_BY_GEONAME_ID = {
            c.geoname_id: c
            for cities in CITIES_BY_COUNTRY.values()
            for c in cities.values()
        }



import dataclasses
import fastapi
import fastapi.encoders
import fastapi.templating
import importlib.resources
import logging
import math
import typing


from .. import dependencies
from ... import geo
from ... import models
from ... import engine
from ... import scoring

LOG = logging.getLogger()


def jsonable(obj: typing.Any) -> typing.Any:
    """Useful filter for Jinja templates: `{{ data | jsonable | tojson }}`"""
    return fastapi.encoders.jsonable_encoder(obj)


def country_with_flag(country_name: str) -> str:
    if not country_name:
        return ""
    return geo.COUNTRIES_BY_NAME[country_name].flag + " " + country_name


def __init_templates() -> fastapi.templating.Jinja2Templates:
    """Initialize Jinja2 templates engine"""
    with importlib.resources.path("archon", "templates") as templates:
        templates = fastapi.templating.Jinja2Templates(
            directory=templates, extensions=["jinja2.ext.i18n"]
        )
        templates.env.filters["jsonable"] = jsonable
        templates.env.filters["country_with_flag"] = country_with_flag
        return templates


TEMPLATES = __init_templates()


router = fastapi.APIRouter(
    default_response_class=fastapi.responses.HTMLResponse,
)


@router.get(
    "/auth/oauth",
    summary="Get an authorization code with the user's approval",
    tags=["oauth"],
)
async def html_auth_oauth(
    client_id: typing.Annotated[str, fastapi.Query()],
    redirect_uri: typing.Annotated[str, fastapi.Query()],
    state: typing.Annotated[str, fastapi.Query()],
    member_uid: dependencies.MemberUidFromSession,
):
    # TODO: display a page asking for the user's authorization
    next = fastapi.datastructures.URL(redirect_uri)
    next = next.include_query_params(
        state=state, code=dependencies.create_authorization_code(client_id, member_uid)
    )
    return fastapi.responses.RedirectResponse(next)


@router.post(
    "/auth/oauth/token",
    summary="Use the authorization code to get a bearer token to use the API.",
    response_class=fastapi.responses.ORJSONResponse,
    tags=["oauth"],
)
async def html_auth_oauth_token(
    grant_type: typing.Annotated[str, fastapi.Form()],
    code: typing.Annotated[str, fastapi.Form()],
    client_uid: dependencies.ClientLogin,
):
    if grant_type != "authorization_code":
        raise fastapi.HTTPException(status_code=403)
    member_uid = dependencies.check_authorization_code(client_uid, code)
    access_token = dependencies.create_access_token(member_uid)
    return dependencies.Token(access_token=access_token, token_type="Bearer")


@router.get(
    "/auth/discord",
    summary="Endpoint where the user is redirected after having logged in Discord",
)
async def html_auth_discord(
    request: fastapi.Request,
    logged_in: dependencies.DiscordLogin,
):
    if logged_in:
        next = request.session.get("next", str(request.url_for("index")))
    else:
        request.session["message"] = "Login failed"
        next = str(request.url_for("login"))
    return fastapi.responses.RedirectResponse(next)


@router.post(
    "/auth/email/reset",
    summary="Ask for a password reset for a given email",
)
async def html_auth_email(
    request: fastapi.Request,
    email: dependencies.EmailAddress,
):
    await dependencies.send_reset_email(email)
    request.session["message"] = (
        "Check your email for the link to create or reset your password."
    )
    return fastapi.responses.RedirectResponse(request.url_for("login"), status_code=303)


@router.get(
    "/auth/email/reset",
    summary="Check the reset link is valid, then ask the user to set a new password",
)
async def html_auth_email(
    request: fastapi.Request,
    uid: dependencies.EmailReset,
):
    if uid:
        return fastapi.responses.RedirectResponse(
            request.url_for("member_display", uid=uid)
        )
    else:
        request.session["message"] = "Email verification URL is invalid or outdated"
        return fastapi.responses.RedirectResponse(request.url_for("login"))


@router.post(
    "/auth/email",
    summary="HTML Basic auth - standard email/password login",
)
async def html_auth_email(
    request: fastapi.Request,
    logged_in: dependencies.EmailLogin,
):
    if logged_in:
        next = request.session.get("next", str(request.url_for("index")))
    else:
        request.session["message"] = (
            "Login failed<br>"
            '<em>Try another login method, or enter your e-mail and click the "Reset Password" button</em>'
        )
        next = str(request.url_for("login"))
    return fastapi.responses.RedirectResponse(next, status_code=303)


@router.get(
    "/auth/token",
    summary="Get a bearer token to use the API.",
    response_class=fastapi.responses.ORJSONResponse,
)
async def html_auth_token(request: fastapi.Request) -> dependencies.Token:
    """Provide an OAuth 2.0 access token for web clients.

    Requires an authenticated session (session cookie).
    This follows the Token-Mediating Backend (TMB) pattern from IETF draft
    "OAuth 2.0 for Browser-Based Applications"
    https://datatracker.ietf.org/doc/html/draft-ietf-oauth-browser-based-apps
    """
    # Not we do not use `dependencies.MemberUidFromSession` as it raise LoginRequired,
    # which ultimately results in an unexpected HTML response here (JSON endpoint)
    if not request.session.get("user_id", None):
        raise fastapi.HTTPException(fastapi.status.HTTP_401_UNAUTHORIZED)
    access_token = dependencies.create_access_token(request.session["user_id"])
    return dependencies.Token(access_token=access_token, token_type="Bearer")


@router.get("/auth/logout/")
async def html_auth_logout(request: fastapi.Request):
    dependencies.anonymous_session(request)
    return fastapi.responses.RedirectResponse(request.url_for("index"))


@router.get("/vekn/claim", response_class=fastapi.responses.HTMLResponse)
async def html_vekn_claim(
    request: fastapi.Request,
    vekn: typing.Annotated[str, fastapi.Query()],
    member_uid: dependencies.MemberUidFromSession,
    op: dependencies.DbOperator,
):
    new_member = await op.claim_vekn(member_uid, vekn)
    if new_member is None:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_403_FORBIDDEN,
            detail="This VEKN does not exist or is already claimed",
        )
    dependencies.authenticated_session(request, new_member)
    dependencies.LOG.warning("uid is now %s", new_member.uid)
    return fastapi.responses.RedirectResponse(
        request.url_for("member_display", uid=new_member.uid)
    )


@router.get("/vekn/abandon", response_class=fastapi.responses.HTMLResponse)
async def html_vekn_abandon(
    request: fastapi.Request,
    member_uid: dependencies.MemberUidFromSession,
    op: dependencies.DbOperator,
):
    new_member = await op.abandon_vekn(member_uid)
    if new_member is None:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_404_NOT_FOUND,
            detail="No VEKN to abandon",
        )
    dependencies.authenticated_session(request, new_member)
    dependencies.LOG.warning("uid is now %s", new_member.uid)
    return fastapi.responses.RedirectResponse(
        request.url_for("member_display", uid=new_member.uid)
    )


@router.get("/")
async def root(request: fastapi.Request):
    return fastapi.responses.RedirectResponse(request.url_for("index"))


@router.get("/index.html")
async def index(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    op: dependencies.DbOperator,
):
    request.session["next"] = str(request.url_for("index"))
    members: list[models.Person] = await op.get_members()
    if "member" not in context:
        for member in members:
            member.name = ""

    # compute rankings
    def ranked(members, category):
        members.sort(key=lambda x: -x.ranking.get(category, 0))
        rank, passed, rating = 0, 0, math.inf
        for member in members:
            member_rating = member.ranking.get(category, 0)
            if member_rating == 0:
                break
            if member_rating < rating:
                rank += 1 + passed
                rating = member_rating
                passed = 0
            else:
                passed += 1
            if rank > 500:
                break
            yield rank, member

    context["members"] = {
        category.value: list(ranked(members, category))
        for category in models.RankingCategoy
    }
    return TEMPLATES.TemplateResponse(
        request=request, name="index.html.j2", context=context
    )


@router.get("/login.html")
async def login(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    next: typing.Annotated[str | None, fastapi.Query()] = None,
):
    if next:
        request.session["next"] = next
    message = request.session.pop("message", None)
    logging.warning("message: %s", message)
    if message:
        context["message"] = message
    return TEMPLATES.TemplateResponse(
        request=request, name="login.html.j2", context=context
    )


@router.get("/tournament/create.html")
async def tournament_create(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    _: dependencies.MemberUidFromSession,
):
    return TEMPLATES.TemplateResponse(
        request=request, name="tournament/edit.html.j2", context=context
    )


@router.get("/tournament/list.html")
async def tournament_list(
    request: fastapi.Request,
    context: dependencies.SessionContext,
):
    request.session["next"] = str(request.url_for("tournament_list"))
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/list.html.j2",
        context=context,
    )


def _filter(cls, filter_cls, obj):
    return cls(**dataclasses.asdict(filter_cls(**dataclasses.asdict(obj))))


@router.get("/tournament/{uid}/display.html")
async def tournament_display(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.Tournament,
):
    request.session["next"] = str(
        request.url_for("tournament_display", uid=tournament.uid)
    )
    # filter out other members info
    member_uid = request.session.get("user_id", None)
    provide_score = set()
    if member_uid:
        provide_score.add(member_uid)
    if tournament.standings_mode == models.StandingsMode.TOP_10:
        for rank, player in engine.standings(tournament):
            if rank > 10:
                break
            provide_score.add(player.uid)
    if tournament.standings_mode == models.StandingsMode.CUTOFF:
        cutoff = scoring.Score()
        for rank, player in engine.standings(tournament):
            cutoff = player.result
            if rank > 5:
                break
        context["cutoff"] = cutoff
    for k, v in tournament.players.items():
        if (
            tournament.standings_mode == models.StandingsMode.PUBLIC
            or tournament.state
            in [models.TournamentState.FINALS, models.TournamentState.FINISHED]
            or k in provide_score
        ):
            tournament.players[k] = _filter(models.Player, models.PlayerInfo, v)
        elif k == member_uid:
            pass
        else:
            tournament.players[k] = _filter(models.Player, models.PublicPerson, v)
    # multideck tournaments: keep the deck info per round only for the player asking
    for round_ in tournament.rounds:
        for table in round_.tables:
            for i, seat in enumerate(table.seating):
                if seat.player_uid != member_uid:
                    table.seating[i] = _filter(models.TableSeat, models.SeatInfo, seat)
    context["tournament"] = tournament
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/display.html.j2",
        context=context,
    )


@router.get("/tournament/{uid}/edit.html")
async def tournament_edit(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.TournamentConfig,
    member_uid: dependencies.MemberUidFromSession,
):
    if member_uid not in [j.uid for j in tournament.judges]:
        raise fastapi.HTTPException(
            status_code=fastapi.status.HTTP_403_FORBIDDEN,
            detail="A judge is required",
        )
    context["tournament"] = tournament
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/edit.html.j2",
        context=context,
    )


@router.get("/tournament/{uid}/console.html")
async def tournament_console(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.Tournament,
    member: dependencies.PersonFromSession,
):
    dependencies.check_can_admin_tournament(member, tournament)
    context["tournament"] = tournament
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/console.html.j2",
        context=context,
    )


@router.get("/tournament/{uid}/checkin.html")
async def tournament_display(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.Tournament,
    member: dependencies.PersonFromSession,
):
    dependencies.check_can_admin_tournament(member, tournament)
    context["name"] = tournament.name
    context["code"] = tournament.checkin_code
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/checkin.html.j2",
        context=context,
    )


@router.get("/tournament/{uid}/print-seating.html")
async def tournament_print_seating(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.Tournament,
    round: typing.Annotated[int, fastapi.Query()],
    member: dependencies.PersonFromSession,
):
    dependencies.check_can_admin_tournament(member, tournament)
    context["round_number"] = round
    context["round"] = []
    if round == len(tournament.rounds) and tournament.finals_seeds:
        context["finals"] = True
        data = []
        for uid in tournament.finals_seeds:
            player = tournament.players[uid]
            data.append({"vekn": player.vekn, "name": player.name})
        context["round"].append(data)
    else:
        for table in tournament.rounds[round - 1].tables:
            data = []
            for seat in table.seating:
                player = tournament.players[seat.player_uid]
                data.append({"vekn": player.vekn, "name": player.name})
            context["round"].append(data)
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/print-seating.html.j2",
        context=context,
    )


@router.get("/tournament/{uid}/print-standings.html")
async def tournament_print_seating(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    tournament: dependencies.Tournament,
    member: dependencies.PersonFromSession,
):
    dependencies.check_can_admin_tournament(member, tournament)
    context["round_number"] = len(tournament.rounds)
    standings = engine.standings(tournament)
    finished = tournament.state == models.TournamentState.FINISHED
    context["finished"] = finished
    if finished:
        context["standings"] = standings
    else:
        match tournament.standings_mode:
            case models.StandingsMode.PRIVATE:
                context["private"] = True
            case models.StandingsMode.CUTOFF:
                context["cutoff"] = standings[4][1].result
            case models.StandingsMode.TOP_10:
                context["standings"] = [(r, p) for r, p in standings if r <= 10]
            case models.StandingsMode.PUBLIC:
                context["standings"] = standings
    return TEMPLATES.TemplateResponse(
        request=request,
        name="tournament/print-standings.html.j2",
        context=context,
    )


@router.get("/document/archon-help.html")
async def document_archon_help(
    request: fastapi.Request, context: dependencies.SessionContext
):
    return TEMPLATES.TemplateResponse(
        request=request,
        name="document/archon-help.html.j2",
        context=context,
    )


@router.get("/document/tournament-rules.html")
async def document_tournament_rules(
    request: fastapi.Request, context: dependencies.SessionContext
):
    return TEMPLATES.TemplateResponse(
        request=request,
        name="document/tournament-rules.html.j2",
        context=context,
    )


@router.get("/document/judges-guide.html")
async def document_judges_guide(
    request: fastapi.Request, context: dependencies.SessionContext
):
    return TEMPLATES.TemplateResponse(
        request=request,
        name="document/judges-guide.html.j2",
        context=context,
    )


@router.get("/document/code-of-ethics.html")
async def document_code_of_ethics(
    request: fastapi.Request, context: dependencies.SessionContext
):
    return TEMPLATES.TemplateResponse(
        request=request,
        name="document/code-of-ethics.html.j2",
        context=context,
    )


@router.get("/member/list.html")
async def member_list(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    member: dependencies.PersonFromSession,
):
    return TEMPLATES.TemplateResponse(
        request=request,
        name="member/list.html.j2",
        context=context,
    )


@router.get("/member/{uid}/display.html")
async def member_display(
    request: fastapi.Request,
    context: dependencies.SessionContext,
    member: dependencies.PersonFromSession,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
):
    if member.uid != uid and not member.vekn:
        target = await op.get_member(uid)
        dependencies.check_can_contact(member, target)
    return TEMPLATES.TemplateResponse(
        request=request,
        name="member/display.html.j2",
        context=context,
    )



import datetime
import enum
import uuid
import pydantic
import secrets
from pydantic import dataclasses

from . import events
from . import scoring


class TournamentFormat(enum.StrEnum):
    Standard = "Standard"
    Limited = "Limited"
    Draft = "Draft"


class TournamentRank(enum.StrEnum):
    BASIC = ""
    NC = "National Championship"
    CC = "Continental Championship"
    GP = "Grand Prix"


class TournamentState(enum.StrEnum):
    REGISTRATION = "Registration"
    WAITING = "Waiting"
    PLAYING = "Playing"
    FINALS = "Finals"
    FINISHED = "Finished"


class StandingsMode(enum.StrEnum):
    PRIVATE = "Private"  # Default
    CUTOFF = "Cutoff"  # Cutoff to make top 5
    TOP_10 = "Top 10"  # Top 10 players
    PUBLIC = "Public"  # All players


class PlayerState(enum.StrEnum):
    REGISTERED = "Registered"
    CHECKED_IN = "Checked-in"
    PLAYING = "Playing"
    FINISHED = "Finished"


class TableState(enum.StrEnum):
    FINISHED = "Finished"
    IN_PROGRESS = "In Progress"
    INVALID = "Invalid"


class Barrier(enum.StrEnum):
    MISSING_DECK = "Missing Deck"
    BANNED = "Banned"
    DISQUALIFIED = "Disqualified"
    MAX_ROUNDS = "Max Rounds"


class RankingCategoy(enum.StrEnum):
    CONSTRUCTED_ONLINE = "Constructed Online"
    CONSTRUCTED_ONSITE = "Constructed Onsite"
    LIMITED_ONLINE = "Limited Online"
    LIMITED_ONSITE = "Limited Onsite"


class MemberRole(enum.StrEnum):
    ADMIN = "Admin"
    PRINCE = "Prince"
    JUDGE = "Judge"  # Elder judge (event judges don't need to be official judges)
    ANC_JUDGE = "Anc. Judge"  # Ancilla Judge
    NEO_JUDGE = "Neo. Judge"  # Neonate Judge
    NC = "NC"  # National Coordinator
    PTC = "PTC"  # Playtest Coordinator
    PLAYTESTER = "Playtester"
    ETHICS = "Ethics"  # Member of the Ethics Comitee


@dataclasses.dataclass
class PublicPerson:
    name: str
    vekn: str = ""
    uid: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    country: str | None = ""  # country name
    country_flag: str | None = ""  # unicode flag char
    city: str | None = ""  # city name


@dataclasses.dataclass
class Person(PublicPerson):
    roles: list[MemberRole] = pydantic.Field(default_factory=list)
    sponsor: str | None = ""  # useful for organizers, to find their recruits
    sanctions: list["RegisteredSanction"] = pydantic.Field(default_factory=list)
    ranking: dict[RankingCategoy, int] = pydantic.Field(default_factory=dict)

    @pydantic.field_validator("ranking", mode="before")
    @classmethod
    def convert_migration_2025_04(cls, v):
        # TODO: drop after migration
        if "constructed_online" in v:
            v[RankingCategoy.CONSTRUCTED_ONLINE.value] = v["constructed_online"]
            del v["constructed_online"]
        if "constructed_onsite" in v:
            v[RankingCategoy.CONSTRUCTED_ONSITE.value] = v["constructed_onsite"]
            del v["constructed_onsite"]
        if "limited_online" in v:
            v[RankingCategoy.LIMITED_ONLINE.value] = v["limited_online"]
            del v["limited_online"]
        if "limited_onsite" in v:
            v[RankingCategoy.LIMITED_ONSITE.value] = v["limited_onsite"]
            del v["limited_onsite"]
        return v


@dataclasses.dataclass
class KrcgCard:
    id: int
    name: str
    count: int
    comments: str = ""


@dataclasses.dataclass
class KrcgCardsGroup:
    type: str
    count: int
    cards: list[KrcgCard] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class KrcgCrypt:
    count: int
    cards: list[KrcgCard] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class KrcgLibrary:
    count: int
    cards: list[KrcgCardsGroup] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class KrcgDeck:
    crypt: KrcgCrypt
    library: KrcgLibrary
    id: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    vdb_link: str = ""
    event: str = ""
    event_link: str = ""
    place: str = ""
    date: datetime.date | None = None
    tournament_format: str = ""
    players_count: int = 0
    player: str = ""
    score: str = ""
    name: str = ""
    author: str = ""
    comments: str = ""


@dataclasses.dataclass
class PlayerInfo(PublicPerson):
    state: PlayerState = PlayerState.REGISTERED
    rounds_played: int = 0
    table: int = 0  # non-zero when playing
    seat: int = 0  # non-zero when playing
    result: scoring.Score = pydantic.Field(default_factory=scoring.Score)
    seed: int = 0  # Finals seed


@dataclasses.dataclass
class Player(Person):
    state: PlayerState = PlayerState.REGISTERED
    rounds_played: int = 0
    table: int = 0  # non-zero when playing
    seat: int = 0  # non-zero when playing
    result: scoring.Score = pydantic.Field(default_factory=scoring.Score)
    seed: int = 0  # Finals seed
    barriers: list[Barrier] = pydantic.Field(default_factory=list)
    toss: int = 0  # non-zero when draws for seeding finals
    deck: KrcgDeck | None = None  # card ID: card count, default deck (monodeck)

    def __hash__(self):
        return hash(self.vekn)


@dataclasses.dataclass
class SeatInfo:
    player_uid: str
    result: scoring.Score = pydantic.Field(default_factory=scoring.Score)


@dataclasses.dataclass
class TableSeat(SeatInfo):
    deck: KrcgDeck | None = None  # card ID: card count
    judge_uid: str = ""  # Result set by a judge cannot be modified by a player


@dataclasses.dataclass
class ScoreOverride:
    judge: Person
    comment: str = ""


@dataclasses.dataclass
class TableInfo:
    seating: list[SeatInfo]


@dataclasses.dataclass
class Table(TableInfo):
    seating: list[TableSeat]
    state: TableState = TableState.IN_PROGRESS
    override: ScoreOverride | None = None


@dataclasses.dataclass
class RoundInfo:
    tables: list[TableInfo]


@dataclasses.dataclass
class Round:
    tables: list[Table]


@dataclasses.dataclass
class LimitedFormat:
    mono_vampire: bool = False
    mono_clan: bool = False
    storyline: str = ""
    include: list[int] = pydantic.Field(default_factory=list)
    exclude: list[int] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class Sanction:
    judge: Person | None = None  # not set on web interfaces, set before entering DB
    uid: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    level: events.SanctionLevel = events.SanctionLevel.CAUTION
    category: events.SanctionCategory = events.SanctionCategory.NONE
    comment: str = ""


@dataclasses.dataclass
class TournamentMinimal:
    name: str
    format: TournamentFormat
    start: datetime.datetime
    finish: datetime.datetime | None = None
    timezone: str = "UTC"
    uid: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    country: str | None = None
    online: bool = False
    rank: TournamentRank = TournamentRank.BASIC
    state: TournamentState = TournamentState.REGISTRATION


@dataclasses.dataclass
class TournamentConfig(TournamentMinimal):
    judges: list[PublicPerson] = pydantic.Field(default_factory=list)
    venue: str = ""
    venue_url: str = ""
    address: str = ""
    map_url: str = ""
    proxies: bool = False
    multideck: bool = False
    decklist_required: bool = True
    description: str = ""
    standings_mode: StandingsMode = StandingsMode.PRIVATE
    max_rounds: int = 0
    limited: LimitedFormat | None = None


@dataclasses.dataclass
class Tournament(TournamentConfig):
    # active tournament console
    # # For now, rounds[-1] is always current.
    # # This might come back to bite us for staggered tournament,
    # # but let's try to avoid it.
    # current_round: int = 0
    checkin_code: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(16)
    )
    # note the doc is not great for this kind of dict, keys show as "additionalProp".
    # Might be improved in a future version of redoc
    # https://github.com/swagger-api/swagger-ui/pull/9739
    players: dict[str, Player] = pydantic.Field(default_factory=dict)
    finals_seeds: list[str] = pydantic.Field(default_factory=list)
    rounds: list[Round] = pydantic.Field(default_factory=list)
    sanctions: dict[str, list[Sanction]] = pydantic.Field(default_factory=dict)
    winner: str = ""
    extra: dict = pydantic.Field(default_factory=dict)  # third-party data if any


@dataclasses.dataclass
class TournamentInfo(TournamentConfig):
    players: dict[str, PublicPerson] = pydantic.Field(default_factory=dict)
    finals_seeds: list[str] = pydantic.Field(default_factory=list)
    rounds: list[RoundInfo] = pydantic.Field(default_factory=list)
    winner: str = ""


# note: cannot use dataclass as query param
class TournamentFilter(pydantic.BaseModel):
    date: str = ""
    uid: str = ""
    country: str = ""
    online: bool = True
    states: list[TournamentState] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class DeckInfo:
    deck: KrcgDeck
    score: scoring.Score
    winner: bool = False


@dataclasses.dataclass
class TournamentDeckInfo(TournamentConfig):
    decks: list[DeckInfo] = pydantic.Field(default_factory=list)


@dataclasses.dataclass
class Country:
    iso: str  # ISO-3166 alpha-2 country code
    iso3: str  # ISO-3166 alpha-3 country code
    iso_numeric: int  # ISO-3166 numeric country code
    fips: str  # FIPS 2 - letters code
    country: str  # Country name
    flag: str  # Country flag (unicode char)
    capital: str  # Capital name
    continent: str  # Continent 2 - letters code(cf.top - level comment)
    tld: str  # Internet Top - Level Domain, including the dot
    currency_code: str  # ISO 4217 alpha - 3 currency code
    currency_name: str  # Currency name
    phone: str  # Phone prefix
    postal_code_regex: str  # Perl / Python regular expression
    languages: list[str]  # list of IETF language tags
    geoname_id: int  # integer id of record in geonames database


@dataclasses.dataclass
class City:
    geoname_id: int  # integer id of record in geonames database
    unique_name: str  # A name unique in the country (w/ admin divisions when needed)
    name: str  # name of geographical point (utf8) varchar(200)
    ascii_name: str  # name of geographical point in plain ascii characters
    latitude: float  # latitude in decimal degrees (wgs84)
    longitude: float  # longitude in decimal degrees (wgs84)
    feature_class: str  # see http://www.geonames.org/export/codes.html
    feature_code: str  # see http://www.geonames.org/export/codes.html
    country_code: str  # ISO-3166 2-letter country code, 2 characters
    country_name: str  # Country name (matches country.country)
    country_flag: str  # Country flag (unicode char)
    cc2: list[str]  # alternate country codes, ISO-3166 2-letter country codes
    admin1: str  # name of first administrative division (state/region)
    admin2: str  # name of second administrative division (county)
    timezone: str  # iana timezone id
    modification_date: str  # date of last modification in ISO format


@dataclasses.dataclass
class DiscordUser:
    id: str  # the user's id (Discord snowflake)
    username: str  # the user's username, not unique across the platform
    discriminator: str  # the user's Discord-tag
    global_name: str | None  # the user's display name, if it is set.
    email: str | None  # the user's email
    verified: bool | None  # whether the email has been verified
    mfa_enabled: bool | None  # whether the user has two factor enabled
    locale: str | None  # the user's chosen language option
    # we ignore the rest:
    # avatar: str | None  # the user's avatar hash
    # bot: bool | None  # whether the user belongs to an OAuth2 application
    # system: bool | None  # whether the user is an Official Discord System user
    # banner: str | None  # the user's banner hash
    # accent_color: int | None  # the user's banner color
    # flags: int | None  # the flags on a user's account
    # premium_type: int | None  # the type of Nitro subscription on a user's account
    # public_flags: int | None  # the public flags on a user's account
    # avatar_decoration_data: dict | None  # data for the user's avatar decoration


@dataclasses.dataclass(kw_only=True)
class RegisteredSanction(Sanction):
    tournament: TournamentConfig | None = None


@dataclasses.dataclass
class TournamentRating:
    tournament: TournamentMinimal
    size: int = 0
    rounds_played: int = 0
    result: scoring.Score = pydantic.Field(default_factory=scoring.Score)
    rank: int = 0
    rating_points: int = 0


@dataclasses.dataclass
class MemberInfo:
    name: str | None = None
    country: str | None = None  # country name
    city: str | None = None  # city name
    nickname: str | None = None  # player nickname (on social, lackey, etc.)
    email: str | None = None  # the user's email
    whatsapp: str | None = None  # phone


@dataclasses.dataclass
class Member(Person):
    nickname: str | None = None  # player nickname (on social, lackey, etc.)
    email: str | None = None  # the user's email
    verified: bool | None = None  # whether the email has been verified
    discord: DiscordUser | None = None  # Discord data
    password_hash: str = ""
    whatsapp: str | None = None  # phone
    ratings: dict[str, TournamentRating] = pydantic.Field(default_factory=dict)
    prefix: str | None = None  # temporary, to compute sponsors when syncing vekn


@dataclasses.dataclass
class Client:
    name: str
    uid: str | None = None  # UUID assigned by the backend


@dataclasses.dataclass
class RoleParameter:
    role: MemberRole


@dataclasses.dataclass
class VeknParameter:
    vekn: str


@dataclasses.dataclass
class PasswordParameter:
    password: str



"""Used for debug purposes only.
So you can `python -m debugpy launch -m archon.app`, and interface with your IDE.
See https://fastapi.tiangolo.com/tutorial/debugging
"""

import logging
import logging.handlers
import uvicorn
import uvicorn.logging
from . import main

if __debug__:
    log_level = logging.DEBUG
else:
    log_level = logging.INFO
logging.getLogger().setLevel(log_level)
logging.basicConfig(level=log_level)
uvicorn.run(main.app, log_level=log_level)



#!/usr/bin/env python3
import asyncio
import typer
import typing

from . import db
from . import models
from . import vekn


app = typer.Typer()


async def async_reset_db(keep_members) -> None:
    async with db.POOL:
        await db.reset(keep_members)
        await db.init()


@app.command()
def reset_db(
    confirm: typing.Annotated[bool, typer.Option(prompt=True)],
    keep_members: typing.Annotated[bool, typer.Option(prompt=True)],
) -> None:
    """⚠️  Reset the database ⚠️  Removes all data"""
    if confirm:
        asyncio.run(async_reset_db(keep_members))


async def async_list() -> None:
    await db.POOL.open()
    async with db.operator() as op:
        tournaments = await op.get_tournaments()
        for tournament in tournaments:
            print(tournament)
    await db.POOL.close()


@app.command()
def list() -> None:
    """List tournaments"""
    asyncio.run(async_list())


async def get_members() -> None:
    async with db.POOL:
        async with db.operator(autocommit=True) as op:
            rankings = await vekn.get_rankings()
            prefixes_map = {}  # prefix owners
            count = 0
            async for members in vekn.get_members_batches():
                for member in members:
                    member.ranking = rankings.get(member.vekn, {})
                # note insert members modifies the passed members list
                # after this call, all members have the right DB uid and data
                async with op.conn.transaction():
                    await op.insert_members(members)
                for member in members:
                    if member.prefix and len(member.prefix) == 3:
                        if member.prefix in prefixes_map:
                            assert prefixes_map[member.prefix].vekn == member.vekn
                        prefixes_map[member.prefix] = member
                    count += 1
                    if not count % 100:
                        print(f" {count}", end="\r", flush=True)
            print("Set sponsors...")
            for prefix, owner in prefixes_map.items():
                async with op.conn.transaction():
                    await op.set_sponsor_on_prefix(prefix, owner.uid)
            print(f"\rDone, {count} members synced")


@app.command()
def sync_members() -> None:
    """Update members from the vekn.net website"""
    asyncio.run(get_members())


async def get_events() -> None:
    async with db.POOL:
        async with db.operator(autocommit=True) as op:
            members = await op.get_members()
            count = 0
            async for event in vekn.get_events(members):
                async with op.conn.transaction():
                    await op.upsert_vekn_tournament(event)
                count += 1
                if not count % 10:
                    print(f" {count}", end="\r", flush=True)
            print(f"\rDone, {count} events synced")


@app.command()
def sync_events() -> None:
    """Update historical tournaments from the vekn.net website"""
    asyncio.run(get_events())


async def db_purge() -> int:
    async with db.POOL:
        async with db.operator() as op:
            return await op.purge_tournament_events()


@app.command()
def purge() -> None:
    """Purge deprecated historical data"""
    count = asyncio.run(db_purge())
    print(f"{count} record{'s' if count > 1 else ''} deleted")


async def async_add_client(name: str) -> tuple[str, str]:
    async with db.POOL:
        async with db.operator() as op:
            client_id = await op.create_client(models.Client(name))
            client_secret = await op.reset_client_secret(client_id)
            return client_id, client_secret


@app.command()
def add_client(name: typing.Annotated[str, typer.Option(prompt=True)]) -> None:
    """Add an authorized client to the platform"""
    client_id, client_secret = asyncio.run(async_add_client(name))
    print("Store the secret safely: if lost, it cannot be retrieved.")
    print(f'CLIENT_ID="{client_id}"')
    print(f'CLIENT_SECRET="{client_secret}"')


async def async_recompute_ratings() -> None:
    async with db.POOL:
        async with db.operator(autocommit=True) as op:
            await op.recompute_all_ratings()


@app.command()
def recompute_ratings() -> None:
    """Recompute all tournament ratings"""
    asyncio.run(async_recompute_ratings())


if __name__ == "__main__":
    app()



import base64
import collections
import contextlib
import dataclasses
import datetime
import dotenv
import hmac
import logging
import orjson
import os
import psycopg
import psycopg.types.json
import psycopg_pool
import secrets
import typing
import uuid
import zoneinfo

from . import events
from . import geo
from . import models
from . import engine

LOG = logging.getLogger()

dotenv.load_dotenv()
DB_USER = os.getenv("DB_USER", "archon")
DB_PWD = os.getenv("DB_PWD", "")
CONNINFO = f"postgresql://{DB_USER}:{DB_PWD}@localhost/archondb"
HASH_KEY = base64.b64decode(os.getenv("HASH_KEY", ""))


psycopg.types.json.set_json_dumps(
    lambda o: orjson.dumps(o, option=orjson.OPT_NON_STR_KEYS)
)
psycopg.types.json.set_json_loads(orjson.loads)


def reconnect_failed(_pool: psycopg_pool.AsyncConnectionPool):
    LOG.error("Failed to reconnect to the PostgreSQL database")


#: await POOL.open() before using this module, and POOL.close() when finished
POOL = psycopg_pool.AsyncConnectionPool(
    CONNINFO,
    open=False,
    max_size=10,
    reconnect_failed=reconnect_failed,
)


class IndexError(RuntimeError): ...


class NotFound(RuntimeError): ...


async def init():
    """Idempotent DB initialization"""
    async with POOL.connection() as conn:
        await conn.set_autocommit(True)
        async with conn.cursor() as cursor:
            LOG.debug("Initialising DB")
            await cursor.execute(
                "CREATE TABLE IF NOT EXISTS members("
                "uid UUID DEFAULT gen_random_uuid() PRIMARY KEY, "
                "vekn TEXT DEFAULT '', "
                "data jsonb)"
            )
            # Unique index on discord ID
            await cursor.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_member_discord_id "
                "ON members "
                "USING BTREE ((data -> 'discord' ->> 'id'))"
            )
            # Unique index on email
            await cursor.execute(
                "CREATE UNIQUE INDEX IF NOT EXISTS idx_member_email "
                "ON members "
                "USING BTREE ((data ->> 'email'))"
            )
            # Index vekn ID#
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_member_vekn "
                "ON members "
                "USING BTREE (vekn)"
            )
            # Index roles
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_member_roles "
                "ON members "
                "USING GIN ((data -> 'roles'))"
            )
            # Trigram index for member names, for quick completion
            await cursor.execute("CREATE EXTENSION IF NOT EXISTS pg_trgm")
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_member_name_trgm "
                "ON members "
                "USING GIST ((data ->> 'name') gist_trgm_ops)"
            )
            await cursor.execute(
                "CREATE TABLE IF NOT EXISTS tournaments("
                "uid UUID DEFAULT gen_random_uuid() PRIMARY KEY, "
                "data jsonb)"
            )
            # Index on tournaments JSON keys and paths
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_tournament_json "
                "ON tournaments "
                "USING GIN (data jsonb_path_ops)"
            )
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_tournament_players "
                "ON tournaments "
                "USING GIN ((data->'players'))"
            )
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_tournament_vekn "
                "ON tournaments "
                "USING BTREE ((data->'extra'->>'vekn_id'))"
            )
            # timetz function to help index tournaments by date
            await cursor.execute(
                "CREATE OR REPLACE FUNCTION timetz(text, text) RETURNS timestamptz "
                "AS $$select ($1 || ' ' || $2)::timestamptz$$ "
                "LANGUAGE SQL IMMUTABLE RETURNS NULL ON NULL INPUT;"
            )
            await cursor.execute(
                "CREATE INDEX IF NOT EXISTS idx_tournament_start "
                "ON tournaments "
                "USING BTREE ((timetz(data ->> 'start', data ->> 'timezone')), uid);"
            )
            await cursor.execute(
                "CREATE TABLE IF NOT EXISTS clients("
                "uid UUID DEFAULT gen_random_uuid() PRIMARY KEY, "
                "secret_hash BYTEA,"
                "data jsonb)"
            )
            await cursor.execute(
                "CREATE TABLE IF NOT EXISTS tournament_events("
                "uid UUID PRIMARY KEY, "
                "timestamp TIMESTAMP WITH TIME ZONE NOT NULL, "
                "tournament_uid UUID REFERENCES tournaments(uid) ON DELETE CASCADE, "
                "member_uid UUID REFERENCES members(uid) ON DELETE SET NULL, "
                "data jsonb)"
            )
        await conn.set_autocommit(False)


class IsDataclass(typing.Protocol):
    # https://stackoverflow.com/a/55240861
    __dataclass_fields__: typing.ClassVar[dict[str, typing.Any]]


def jsonize(datacls: IsDataclass):
    return psycopg.types.json.Jsonb(dataclasses.asdict(datacls))


async def reset(keep_members: bool = True):
    """ONLY FROM CLI - LOSES ALL DATA"""
    async with POOL.connection() as conn:
        async with conn.cursor() as cursor:
            LOG.warning("Reset DB")
            await cursor.execute("DROP TABLE IF EXISTS tournament_events")
            await cursor.execute("DROP TABLE IF EXISTS tournaments")
            if not keep_members:
                await cursor.execute("DROP TABLE IF EXISTS members")


T = typing.TypeVar("T", bound=models.Tournament)


@contextlib.asynccontextmanager
async def member_consistency() -> typing.AsyncIterator[None]:
    try:
        yield
    except psycopg.errors.IntegrityError as err:
        if err.diag.constraint_name == "idx_member_email":
            raise IndexError("This email already exists")
        if err.diag.constraint_name == "idx_member_discord_id":
            raise IndexError("Discord ID already taken")
        raise


class Operator:
    """All async operations available for the API layer"""

    def __init__(self, conn: psycopg.AsyncConnection):
        self.conn = conn

    async def create_client(self, client: models.Client) -> str:
        """Create a client, returns its uid"""
        client.uid = client.uid or str(uuid.uuid4())
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "INSERT INTO clients (uid, data) " "VALUES (%s, %s) RETURNING uid",
                [client.uid, jsonize(client)],
            )
            return str((await res.fetchone())[0])

    async def reset_client_secret(self, client_uid: str) -> str:
        """Reset a client secret, store its hash, return the secret."""
        secret = secrets.token_urlsafe(32)
        if not HASH_KEY:
            raise RuntimeError("No HASH_KEY provided by the environment")
        secret_hash = hmac.digest(
            HASH_KEY, base64.urlsafe_b64decode(secret + "=="), "sha3_512"
        )
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "UPDATE clients SET secret_hash=%s WHERE uid=%s",
                [secret_hash, client_uid],
            )
            if res.rowcount < 1:
                raise KeyError(f"Client {client_uid} not found")
            return secret

    async def check_client_secret(self, client_uid: str, secret: str) -> bool:
        if not HASH_KEY:
            raise RuntimeError("No HASH_KEY provided by the environment")
        secret_hash = hmac.digest(
            HASH_KEY, base64.urlsafe_b64decode(secret + "=="), "sha3_512"
        )
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT secret_hash FROM clients WHERE uid=%s",
                [client_uid],
            )
            stored_hash = (await res.fetchone())[0]
        return hmac.compare_digest(secret_hash, stored_hash)

    async def create_tournament(self, tournament: models.TournamentConfig) -> str:
        """Create a tournament, returns its uid"""
        uid = uuid.uuid4()
        tournament.uid = str(uid)
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "INSERT INTO tournaments (uid, data) VALUES (%s, %s) RETURNING uid",
                [uid, jsonize(tournament)],
            )
            return str((await res.fetchone())[0])

    async def upsert_vekn_tournament(self, tournament: models.Tournament) -> str:
        """Insert or update a VEKN tournament.
        Note if it was ours to begin with (and synced down to VEKN when finished),
        we don't sync anything from VEKN.
        """
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM tournaments WHERE data->'extra'->>'vekn_id' = %s",
                [tournament.extra.get("vekn_id")],
            )
            data = await res.fetchone()
            if data:
                if not data[0]["extra"].get("external"):
                    return
                uid = uuid.UUID(data[0]["uid"])
                tournament.uid = str(uid)
                tournament.extra["external"] = True
                res = await cursor.execute(
                    "UPDATE tournaments SET data=%s WHERE uid=%s",
                    [jsonize(tournament), uid],
                )
                return str(uid)
            else:
                uid = uuid.uuid4()
                tournament.uid = str(uid)
                tournament.extra["external"] = True
                res = await cursor.execute(
                    "INSERT INTO tournaments (uid, data) "
                    "VALUES (%s, %s) RETURNING uid",
                    [uid, jsonize(tournament)],
                )
                return str((await res.fetchone())[0])

    async def get_tournaments(
        self, cls: typing.Type[T] = models.Tournament
    ) -> list[models.Tournament]:
        """List all tournaments.
        TODO: paginate. We'll need an index on tournament.start
        """
        async with self.conn.cursor() as cursor:
            res = await cursor.execute("SELECT data FROM tournaments")
            return [cls(**row[0]) for row in await res.fetchall()]

    async def get_tournaments_minimal(
        self,
        filter: models.TournamentFilter | None = None,
    ) -> tuple[models.TournamentFilter, list[models.TournamentMinimal]]:
        """List all tournaments with minimal information"""
        Q = (
            "SELECT "
            "data ->> 'name', "
            "data ->> 'format', "
            "data ->> 'start', "
            "data ->> 'finish', "
            "data ->> 'timezone', "
            "uid::text, "
            "data ->> 'country', "
            "data ->> 'online', "
            "data ->> 'rank', "
            "data ->> 'state' "
            "FROM tournaments "
        )
        pieces = []
        args = []
        if filter:
            Q += "WHERE "
        # Note: bad practice to use OFFSET, it becomes inefficient the larger it is
        # using cursor in the WHERE clause with the appropriate index is the way to go.
        if filter:
            if filter.uid:
                pieces.append(
                    "(timetz(data ->> 'start', data ->> 'timezone') < %s::timestamptz "
                    "OR ("
                    "timetz(data ->> 'start', data ->> 'timezone') = %s::timestamptz "
                    "AND uid < %s))"
                )
                args.extend([filter.date, filter.date, uuid.UUID(filter.uid)])
            if filter.country:
                pieces.append(
                    "(data ->> 'country' IS NULL OR data ->> 'country' IN ('', %s))"
                )
                args.append(filter.country)
            if not filter.online:
                pieces.append("(data ->> 'online')::boolean IS FALSE")
            if filter.states:
                pieces.append("data ->> 'state' = ANY(%s)")
                args.append(filter.states)
        Q += " AND ".join(pieces)
        Q += (
            " ORDER BY timetz(data ->> 'start', data ->> 'timezone') DESC, uid DESC "
            "LIMIT 100"
        )
        async with self.conn.cursor() as cursor:
            LOG.debug("Query: %s", Q)
            res = await cursor.execute(Q, args)
            ret = [models.TournamentMinimal(*row) for row in await res.fetchall()]
            ret_filter = models.TournamentFilter()
            if filter:
                ret_filter.country = filter.country
                ret_filter.online = filter.online
                ret_filter.states = filter.states
            if ret and len(ret) == 100:
                ret_filter.date = ret[-1].start.isoformat() + " " + ret[-1].timezone
                ret_filter.uid = ret[-1].uid
            return (ret_filter, ret)

    async def get_tournament(
        self, uid: str, cls: typing.Type[T] = models.Tournament
    ) -> T:
        """Get a tournament by its uid"""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM tournaments WHERE uid=%s", [uuid.UUID(uid)]
            )
            data = await res.fetchone()
            if not data:
                return None
            return cls(**(data[0]))

    async def played_tournaments(self, uid: str) -> list[models.Tournament]:
        """Get all tournament played by a given member"""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM tournaments WHERE data->'players' ? %s", [uid]
            )
            return [models.Tournament(**data[0]) for data in await res.fetchall()]

    async def update_tournament(self, tournament: models.Tournament) -> str:
        """Update a tournament, returns its uid"""
        uid = uuid.UUID(tournament.uid)
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "UPDATE tournaments SET data=%s WHERE uid=%s",
                [jsonize(tournament), uid],
            )
            if res.rowcount < 1:
                raise KeyError(f"Tournament {uid} not found")
            if tournament.state == models.TournamentState.FINISHED:
                await cursor.executemany(
                    f"""UPDATE members 
                    SET data=jsonb_set(data, '{{ratings,{tournament.uid}}}', %s, true) 
                    WHERE uid=%s
                    """,
                    [
                        [jsonize(rating), uuid.UUID(uid)]
                        for uid, rating in engine.ratings(tournament).items()
                    ],
                )
            return str(uid)

    async def delete_tournament(self, uid: str) -> None:
        """Delete a tournament"""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "DELETE FROM tournaments WHERE uid=%s", [uuid.UUID(uid)]
            )
            if res.rowcount < 1:
                raise KeyError(f"Tournament {uid} not found")

    async def record_event(
        self, tournament_uid: str, member_uid: str, event: events.TournamentEvent
    ) -> None:
        tournament_uid = uuid.UUID(tournament_uid)
        member_uid = uuid.UUID(member_uid)
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        async with self.conn.cursor() as cursor:
            await cursor.execute(
                "INSERT INTO tournament_events VALUES (%s, %s, %s, %s, %s)",
                [event.uid, timestamp, tournament_uid, member_uid, jsonize(event)],
            )
            if cursor.rowcount < 1:
                raise RuntimeError("INSERT failed")

    async def purge_tournament_events(self) -> int:
        cutoff = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
            days=366
        )
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "DELETE FROM tournament_events WHERE timestamp < %s", [cutoff]
            )
            return res.rowcount

    async def insert_members(self, members: list[models.Member]) -> None:
        """Insert members from VEKN.
        Modify the members list in place to match DB (ie. uuids).
        """
        async with self.conn.cursor() as cursor:
            # get existing VEKN, lock the table
            res = await cursor.execute(
                "SELECT vekn, data FROM members WHERE vekn=ANY(%s) FOR UPDATE",
                [[m.vekn for m in members]],
            )
            existing = await res.fetchall()
            existing = {e[0]: self._instanciate_member(e[1]) for e in existing}
            # do not overwrite local data & lose relevant info (sanctions, login, etc.)
            # don't revert changes on name
            # TODO: also don't revert country, city, roles and sponsor changes
            # once we're stable on this
            for i, m in enumerate(members):
                if m.vekn in existing:
                    local: models.Member = existing[m.vekn]
                    # TODO remove country & city
                    local.country = m.country
                    local.city = m.city
                    local.ranking = m.ranking
                    local.roles = m.roles
                    local.prefix = m.prefix
                    local.sponsor = None
                    members[i] = local
            # update existing
            # cannot run two prepared statements in parallel, just wait
            await cursor.executemany(
                "UPDATE members SET data=%s WHERE vekn=%s",
                [
                    [self._jsonize_member(m), m.vekn]
                    for m in members
                    if m.vekn in existing
                ],
            )
            # insert new
            await cursor.executemany(
                "INSERT INTO members (uid, vekn, data) VALUES (%s, %s, %s)",
                [
                    [uuid.UUID(m.uid), m.vekn, self._jsonize_member(m)]
                    for m in members
                    if m.vekn not in existing
                ],
            )

    async def insert_member(self, member: models.Member) -> models.Member:
        """Insert a new member"""
        # vekn must not be set
        member.vekn = ""
        async with self.conn.cursor() as cursor, member_consistency():
            # insert new
            await cursor.execute(
                "INSERT INTO members (uid, data) VALUES (%s, %s)",
                [uuid.UUID(member.uid), self._jsonize_member(member)],
            )
            if cursor.rowcount < 1:
                raise RuntimeError("INSERT failed")
            return member

    async def get_member(
        self, uid: str, for_update=False, cls: typing.Type[T] = models.Member
    ) -> T:
        """Get a member from their uid"""
        async with self.conn.cursor() as cursor:
            if for_update:
                query = "SELECT data FROM members WHERE uid=%s FOR UPDATE"
            else:
                query = "SELECT data FROM members WHERE uid=%s"
            res = await cursor.execute(query, [uuid.UUID(uid)])
            data = await res.fetchone()
            if data:
                return self._instanciate_member(data[0], cls)
            return None

    async def get_members(self, uids: list[str] | None = None) -> list[models.Person]:
        """Get all members"""
        async with self.conn.cursor() as cursor:
            if uids:
                res = await cursor.execute(
                    "SELECT data FROM members WHERE uid = ANY(%s)", [uids]
                )
            else:
                res = await cursor.execute("SELECT data FROM members")
            return [
                self._instanciate_member(data[0], models.Person)
                for data in await res.fetchall()
            ]

    async def get_externally_visible_members(
        self, user: models.Person
    ) -> list[models.Person]:
        """Get all members that can be contacted by non-vekn members"""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM members "
                "WHERE data->'roles' ? 'NC' OR data->'roles' ? 'Prince' OR uid=%s",
                [user.uid],
            )
            return [
                self._instanciate_member(data[0], models.Person)
                for data in await res.fetchall()
            ]

    def _update_discord_data(self, member: models.Member, user: models.DiscordUser):
        member.discord = user
        member.name = member.name or user.global_name or user.username
        member.nickname = member.nickname or user.global_name or user.username
        # do not update the email by default
        member.email = member.email or user.email
        if member.email == user.email:
            member.verified = user.verified

    async def update_member_discord(
        self, member_uid: str, user: models.DiscordUser
    ) -> models.Member:
        """Update an existing member, set their discord profile"""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM members WHERE uid = %s FOR UPDATE",
                [uuid.UUID(member_uid)],
            )
            data = await res.fetchone()
            if not data:
                raise RuntimeError(f"Unknown member {member_uid}")
            member = self._instanciate_member(data[0])
            self._update_discord_data(member, user)
            await cursor.execute(
                "UPDATE members SET data=%s WHERE uid = %s",
                [self._jsonize_member(member), uuid.UUID(member_uid)],
            )
            if cursor.rowcount < 1:
                raise RuntimeError(
                    f"Failed to update discord info for member {member_uid}"
                )
            return member

    async def upsert_member_discord(self, user: models.DiscordUser) -> models.Member:
        """Get or create a member from their discord profile"""
        async with self.conn.cursor() as cursor:
            # if this discord ID already exists, use it
            res = await cursor.execute(
                "SELECT data FROM members "
                "WHERE data -> 'discord' ->> 'id' = %s FOR UPDATE",
                [user.id],
            )
            data = await res.fetchone()
            if data:
                member = self._instanciate_member(data[0])
                self._update_discord_data(member, user)
                await cursor.execute(
                    "UPDATE members SET data=%s WHERE data -> 'discord' ->> 'id' = %s",
                    [self._jsonize_member(member), user.id],
                )
                if cursor.rowcount < 1:
                    raise RuntimeError(f"Failed to update Discord ID# {user.id}")
                return member
            # if the email exists already, use it
            if user.email:
                res = await cursor.execute(
                    "SELECT data FROM members "
                    "WHERE data ->> 'email' = %s FOR UPDATE",
                    [user.email],
                )
                data = await res.fetchone()
                if data:
                    member = self._instanciate_member(data[0])
                    self._update_discord_data(member, user)
                    await cursor.execute(
                        "UPDATE members SET data=%s WHERE data ->> 'email' = %s",
                        [self._jsonize_member(member), user.email],
                    )
                    if cursor.rowcount < 1:
                        raise RuntimeError(f"Failed to update Discord ID# {user.id}")
                    return member
            # otherwise create a new member
            uid = uuid.uuid4()
            member = models.Member(
                uid=str(uid),
                vekn="",
                name=user.global_name or user.username,
                nickname=user.global_name or user.username,
                email=user.email,
                verified=user.verified,
                discord=user,
            )
            await cursor.execute(
                "INSERT INTO members (uid, vekn, data) VALUES (%s, %s, %s)",
                [uid, member.vekn, self._jsonize_member(member)],
            )
            if cursor.rowcount < 1:
                raise RuntimeError("INSERT failed")
            return member

    async def upsert_member_email(self, email: str) -> models.Member:
        async with self.conn.cursor() as cursor, member_consistency():
            res = await cursor.execute(
                "SELECT data FROM members " "WHERE data ->> 'email' = %s FOR UPDATE",
                [email],
            )
            data = await res.fetchone()
            if data:
                member = self._instanciate_member(data[0])
                member.verified = True
                await cursor.execute(
                    "UPDATE members SET data=%s WHERE data ->> 'email' = %s",
                    [self._jsonize_member(member), email],
                )
                if cursor.rowcount < 1:
                    raise RuntimeError(f"Failed to update email {email}")
            else:
                uid = uuid.uuid4()
                member = models.Member(
                    uid=str(uid),
                    vekn="",
                    name="",
                    nickname="",
                    email=email,
                    verified=True,
                )
                await cursor.execute(
                    "INSERT INTO members (uid, vekn, data) VALUES (%s, %s, %s)",
                    [uid, member.vekn, self._jsonize_member(member)],
                )
                if cursor.rowcount < 1:
                    raise RuntimeError("INSERT failed")
            return member

    async def get_member_by_email(self, email: str) -> models.Member:
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM members " "WHERE data ->> 'email' = %s",
                [email],
            )
            data = await res.fetchone()
            if data:
                member = self._instanciate_member(data[0])
                return member
            raise NotFound(f"Member not found by email: {email}")

    async def update_member(self, member: models.Member) -> models.Member:
        async with self.conn.cursor() as cursor, member_consistency():
            await cursor.execute(
                "UPDATE members SET data=%s WHERE uid=%s",
                [self._jsonize_member(member), member.uid],
            )
            if cursor.rowcount < 1:
                raise RuntimeError(f"Failed to update member {member.uid}")
            return member

    async def update_member_new_vekn(self, member: models.Member) -> models.Member:
        """Use an atomic query with subquery to set a new VEKN

        This avoid concurrency issues. Also update member data (usually set the sponsor)
        """
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                """UPDATE members
                SET vekn=next_vekn.n
                FROM (
                    SELECT n from generate_series(1000001, 9999999) as n
                    WHERE n NOT IN (select vekn::int from members where vekn <> '')
                    LIMIT 1
                ) as next_vekn 
                WHERE uid=%s
                RETURNING vekn
                """,
                [uuid.UUID(member.uid)],
            )
            data = await res.fetchone()
            if not data:
                raise RuntimeError(f"Failed to assign a new VEKN to {member.uid}")
            member.vekn = data[0]
            await cursor.execute(
                "UPDATE members SET data=%s WHERE uid=%s",
                [self._jsonize_member(member), member.uid],
            )
            if cursor.rowcount < 1:
                raise RuntimeError(f"Failed to update member {member.uid}")
            # also set the vekn in all tournaments the player has been in
            await cursor.execute(
                f"""UPDATE tournaments
                SET data=jsonb_set(data, '{{players,{member.uid},vekn}}', %s, false)
                WHERE data->'players' ? %s
                """,
                [psycopg.types.json.Jsonb(member.vekn), member.uid],
            )
            return await self.update_member(member)

    async def claim_vekn(self, uid: str, vekn: str) -> models.Member | None:
        """Claim an existing vekn ID: returns a **new uid** for the member."""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT vekn, data FROM members WHERE uid=%s FOR UPDATE",
                [uuid.UUID(uid)],
            )
            data = await res.fetchone()
            if not data:
                return None
            old_vekn = data[0]
            member = self._instanciate_member(data[1])
            if old_vekn == vekn:
                #  no change
                LOG.info("%s claiming %s but already has it", uid, vekn)
                return member
            if old_vekn:
                LOG.warning("%s claiming %s, previously had %s", uid, vekn, old_vekn)
                raise RuntimeError("Cannot claim a VEKN: already has one")
            res = await cursor.execute(
                "SELECT data FROM members WHERE vekn=%s FOR UPDATE", [vekn]
            )
            data = await res.fetchone()
            if not data:
                LOG.warning("no VEKN record found for %s", vekn)
                return None
            vekn_member = self._instanciate_member(data[0])
            if vekn_member.discord:
                if vekn_member.discord.id != member.discord.id:
                    LOG.warning(
                        "VEKN %s already owned by %s", vekn, vekn_member.discord.id
                    )
                    return None
            # assign personal elements
            vekn_member.discord = member.discord
            vekn_member.nickname = member.nickname
            vekn_member.email = member.email
            vekn_member.whatsapp = member.whatsapp
            vekn_member.verified = member.verified
            # delete initial (non-vekn) member
            await cursor.execute("DELETE FROM members WHERE uid=%s", [uid])
            # update the VEKN record
            await cursor.execute(
                "UPDATE members SET data=%s WHERE vekn=%s",
                [self._jsonize_member(vekn_member), vekn],
            )
            if cursor.rowcount < 1:
                raise RuntimeError(f"Failed to update VEKN {vekn}")
            return vekn_member

    async def abandon_vekn(self, uid: str) -> models.Member | None:
        """Abandon vekn ID: returns a **new uid** for the member."""
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT vekn, data FROM members WHERE uid=%s FOR UPDATE",
                [uuid.UUID(uid)],
            )
            data = await res.fetchone()
            if not data:
                LOG.warning("User not found: %s", uid)
                return None
            vekn = data[0]
            member = self._instanciate_member(data[1])
            if not vekn:
                LOG.info("No vekn for %s, nothing to do", uid)
                return member
            new_member = models.Member(**dataclasses.asdict(member))
            new_uid = uuid.uuid4()
            new_member.vekn = ""
            new_member.roles = []
            new_member.sanctions = []
            new_member.ranking = models.Ranking()
            new_member.ratings = {}
            new_member.sponsor = ""
            new_member.uid = str(new_uid)
            member.discord = None
            member.nickname = None
            member.email = None
            member.whatsapp = None
            member.verified = False
            await cursor.execute(
                "UPDATE members SET data=%s WHERE uid=%s",
                [self._jsonize_member(member), uuid.UUID(uid)],
            )
            if cursor.rowcount < 1:
                raise RuntimeError(f"Failed to find Member {uid}")
            LOG.debug("Old member updated: %s", uid)
            if new_member.discord:
                nick = new_member.discord.global_name or new_member.discord.username
                new_member.name = nick
                new_member.nickname = nick
                new_member.email = new_member.discord.email
                new_member.verified = new_member.discord.verified
            await cursor.execute(
                "INSERT INTO members (uid, data) VALUES (%s, %s)",
                [new_uid, self._jsonize_member(new_member)],
            )
            LOG.debug("New member created: %s - %s", new_member.uid, data)
            return new_member

    async def set_sponsor_on_prefix(self, prefix: str, sponsor_uid: str):
        async with self.conn.cursor() as cursor:
            res = await cursor.execute(
                "SELECT data FROM members WHERE vekn LIKE %s FOR UPDATE",
                [prefix.replace("%", "") + "%"],
            )
            recruits = [
                self._instanciate_member(data[0]) for data in await res.fetchall()
            ]
            for recruit in recruits:
                recruit.sponsor = sponsor_uid
            await cursor.executemany(
                "UPDATE members SET data=%s WHERE uid=%s",
                [[self._jsonize_member(m), uuid.UUID(m.uid)] for m in recruits],
            )

    async def recompute_all_ratings(self):
        async with self.conn.cursor() as cursor:
            # first get all the tournaments and compute the ratings for everyone
            res = await cursor.execute(
                """SELECT data FROM tournaments
                WHERE data->>'state' = %s
                """,
                [models.TournamentState.FINISHED],
            )
            all_ratings: dict[str, dict[str, models.TournamentRating]] = (
                collections.defaultdict(dict)
            )
            async for row in res:
                tournament = models.Tournament(**row[0])
                for uid, rating in engine.ratings(tournament).items():
                    all_ratings[uid][tournament.uid] = rating
            # clean reset for all members
            await cursor.execute(
                "UPDATE members SET data=jsonb_set(data, '{ratings}', '{}', true)"
            )
            # set all tournaments ratings
            await cursor.executemany(
                """UPDATE members
                SET data=jsonb_set(data, '{ratings}', %s, false)
                WHERE uid=%s
                """,
                [
                    [
                        psycopg.types.json.Jsonb(
                            {k: dataclasses.asdict(v) for k, v in ratings.items()}
                        ),
                        uuid.UUID(uid),
                    ]
                    for uid, ratings in all_ratings.items()
                ],
            )
            # now compute the current rankings, using results from the past 18 months
            # only the top 8 in each category count
            rankings_lists: dict[str, dict[models.RankingCategoy, list[int]]] = (
                collections.defaultdict(lambda: collections.defaultdict(list))
            )
            # compute cutoff: 18 months before today (UTC) at 00:00
            cutoff = datetime.datetime.now(datetime.timezone.utc)
            cutoff = cutoff.replace(hour=0, minute=0, second=0, microsecond=0)
            cutoff = cutoff.replace(year=cutoff.year - 1)
            if cutoff.month > 6:
                cutoff = cutoff.replace(month=cutoff.month - 6)
            else:
                cutoff = cutoff.replace(year=cutoff.year - 1)
                cutoff = cutoff.replace(month=cutoff.month + 6)
            for uid, ratings in all_ratings.items():
                for rating in ratings.values():
                    if (
                        rating.tournament.start.replace(
                            tzinfo=zoneinfo.ZoneInfo(tournament.timezone)
                        )
                        < cutoff
                    ):
                        continue
                    if rating.tournament.format == models.TournamentFormat.Standard:
                        if rating.tournament.online:
                            category = models.RankingCategoy.CONSTRUCTED_ONLINE
                        else:
                            category = models.RankingCategoy.CONSTRUCTED_ONSITE
                    else:
                        if rating.tournament.online:
                            category = models.RankingCategoy.LIMITED_ONLINE
                        else:
                            category = models.RankingCategoy.LIMITED_ONSITE
                    rankings_lists[uid][category].append(rating.rating_points)
            # take the top 8
            rankings = {
                uid: {
                    category: sum(sorted(r, reverse=True)[:8])
                    for category, r in res.items()
                    if r
                }
                for uid, res in rankings_lists.items()
                if res
            }
            # clean everyone's ratings, then update with the newly computed values
            await cursor.execute(
                "UPDATE members SET data=jsonb_set(data, '{ranking}', '{}', true)"
            )
            await cursor.executemany(
                """UPDATE members
                SET data=jsonb_set(data, '{ranking}', %s, false)
                WHERE uid=%s
                """,
                [
                    [
                        psycopg.types.json.Jsonb(ranking),
                        uuid.UUID(uid),
                    ]
                    for uid, ranking in rankings.items()
                ],
            )

    def _jsonize_member(self, member: models.Member):
        data = dataclasses.asdict(member)
        if member.country:
            country = geo.COUNTRIES_BY_NAME[member.country]
            data["country_flag"] = country.flag
            data["country_iso"] = country.iso
            data["country_geoname_id"] = country.geoname_id
        else:
            member.city = ""
        if member.city:
            data["city_geoname_id"] = geo.CITIES_BY_COUNTRY[member.country][
                member.city
            ].geoname_id
        return psycopg.types.json.Json(data)

    def _instanciate_member(self, data: dict, cls: typing.Type[T] = models.Member) -> T:
        if "country_iso" in data:
            country = geo.COUNTRIES_BY_ISO[data["country_iso"]]
            data["country"] = country.country
            data["country_flag"] = country.flag
        if (
            "city_geoname_id" in data
            and data["city_geoname_id"] in geo.CITIES_BY_GEONAME_ID
        ):
            data["city"] = geo.CITIES_BY_GEONAME_ID[data["city_geoname_id"]].unique_name
        else:
            data["city"] = None
        return cls(**data)


@contextlib.asynccontextmanager
async def operator(autocommit: bool = False) -> typing.AsyncIterator[Operator]:
    """Yields an async DB Operator to execute DB operations in a single transaction"""
    async with POOL.connection() as conn:
        if autocommit:
            await conn.set_autocommit(True)
        yield Operator(conn)
        if autocommit:
            await conn.set_autocommit(False)



import asyncio
import contextlib
import dotenv
import fastapi
import fastapi.responses
import fastapi.staticfiles
import importlib.resources
import logging
import os
import starlette.middleware.sessions
import uvicorn.logging

from .. import db
from .. import engine
from .. import vekn
from . import dependencies
from .api import tournament as api__tournament
from .api import vekn as api__vekn
from .html import website as html__website

dotenv.load_dotenv()
SESSION_KEY = os.getenv("SESSION_KEY", "dev_key")

LOG = logging.getLogger()
handler = logging.StreamHandler()
handler.setFormatter(uvicorn.logging.DefaultFormatter("%(levelprefix)s %(message)s"))
LOG.addHandler(handler)
if __debug__ or os.getenv("DEBUG"):
    LOG.setLevel(logging.DEBUG)
    handler.setLevel(logging.DEBUG)


async def sync_vekn_members(op: db.Operator) -> None:
    rankings = await vekn.get_rankings()
    prefixes_map = {}  # prefix owners
    async for members in vekn.get_members_batches():
        for member in members:
            member.ranking = rankings.get(member.vekn, {})
        # note insert members modifies the passed members list
        # after this call, all members have the right DB uid and data
        async with op.conn.transaction():
            await op.insert_members(members)
        for member in members:
            if member.prefix and len(member.prefix) == 3:
                if member.prefix in prefixes_map:
                    assert prefixes_map[member.prefix].vekn == member.vekn
                prefixes_map[member.prefix] = member
    for prefix, owner in prefixes_map.items():
        async with op.conn.transaction():
            await op.set_sponsor_on_prefix(prefix, owner.uid)


async def sync_vekn() -> int | None:
    if __debug__:
        return 1
    async with db.operator(autocommit=True) as op:
        await op.purge_tournament_events()
        await sync_vekn_members(op)
        members = await op.get_members()
        async for event in vekn.get_events(members):
            async with op.conn.transaction():
                await op.upsert_vekn_tournament(event)
        await op.recompute_all_ratings()


def log_sync_errors(task: asyncio.Task) -> None:
    if task.cancelled():
        LOG.info("VEKN sync cancelled")
        return
    exception = task.exception()
    if exception:
        LOG.exception("VEKN sync failed", exc_info=exception)
        return
    if task.result():
        LOG.info("VEKN sync omitted (debug mode)")
    else:
        LOG.info("VEKN sync done")


@contextlib.asynccontextmanager
async def lifespan(app: fastapi.FastAPI):
    """Initialize the DB pool"""
    LOG.debug("Entering APP lifespan")
    async with db.POOL:
        # idempotent init, call it every time
        await db.init()
        # sync VEKN asynchronously, start in the meantime
        task = asyncio.create_task(sync_vekn())
        task.add_done_callback(log_sync_errors)
        yield
        task.cancel()
    LOG.debug("Exiting APP lifespan")


tags_metadata = [
    {
        "name": "vekn",
        "description": "VEKN profile management. The **login** logic is also here.",
    },
    {
        "name": "tournament",
        "description": "Tournament management.",
    },
]


app = fastapi.FastAPI(lifespan=lifespan, tags_metadata=tags_metadata, debug=__debug__)
app.add_middleware(
    starlette.middleware.sessions.SessionMiddleware,
    secret_key=SESSION_KEY,
)
app.openapi = dependencies.custom_openapi(app)

# mount static files
with (
    importlib.resources.path("archon", "img") as img,
    importlib.resources.path("archon", "static") as static,
):
    app.mount(
        "/img",
        app=fastapi.staticfiles.StaticFiles(directory=img),
        name="img",
    )
    app.mount(
        "/static",
        app=fastapi.staticfiles.StaticFiles(directory=static),
        name="static",
    )


# mount routers
app.include_router(html__website.router)
app.include_router(api__tournament.router)
app.include_router(api__vekn.router)


# login redirection
@app.exception_handler(dependencies.LoginRequired)
def auth_exception_handler(
    request: fastapi.Request, exc: dependencies.LoginRequired
) -> fastapi.responses.HTMLResponse:
    """
    Redirect the user to the login page if not logged in
    """
    LOG.debug("Login failed", exc_info=exc.with_traceback(None))
    request.session["message"] = "You need to log in"
    return fastapi.responses.RedirectResponse(
        url=request.url_for("login").include_query_params(next=str(request.url))
    )


# engine errors display
@app.exception_handler(engine.TournamentError)
def engine_exception_handler(
    request: fastapi.Request, exc: engine.TournamentError
) -> fastapi.responses.JSONResponse:
    """
    Returns a well-serialized JSON error message
    """
    if __debug__:
        LOG.exception("tournament error", exc_info=exc)
    else:
        LOG.warning(exc.args[0])
    return fastapi.responses.JSONResponse({"detail": str(exc)}, 400)


# members consistency errors
@app.exception_handler(db.IndexError)
def engine_exception_handler(
    request: fastapi.Request, exc: engine.TournamentError
) -> fastapi.responses.JSONResponse:
    """
    Returns a well-serialized JSON error message
    """
    if __debug__:
        LOG.exception("member index error", exc_info=exc)
    else:
        LOG.warning(exc.args[0])
    return fastapi.responses.JSONResponse({"detail": exc.args[0]}, 400)



import itertools
import logging
import math
import pydantic
from pydantic import dataclasses

LOG = logging.getLogger()


@dataclasses.dataclass(order=True, eq=True)
class Score:
    gw: int = pydantic.Field(0, ge=0)
    vp: float = pydantic.Field(0.0, ge=0, multiple_of=0.5)
    tp: int = pydantic.Field(0, ge=0)

    def __str__(self):
        if self.gw:
            return f"{self.gw}GW{self.vp:.2g} ({self.tp}TP)"
        else:
            return f"{self.vp:.2g}VP ({self.tp}TP)"

    def __add__(self, rhs):
        return self.__class__(
            gw=self.gw + rhs.gw, vp=self.vp + rhs.vp, tp=self.tp + rhs.tp
        )

    def __iadd__(self, rhs):
        self.gw += rhs.gw
        self.vp += rhs.vp
        self.tp += rhs.tp
        return self

    def __sub__(self, rhs):
        return self.__class__(
            gw=max(0, self.gw - rhs.gw),
            vp=max(0, self.vp - rhs.vp),
            tp=max(0, self.tp - rhs.tp),
        )

    def __isub__(self, rhs):
        self.gw -= min(self.gw, rhs.gw)
        self.vp -= min(self.vp, rhs.vp)
        self.tp -= min(self.tp, rhs.tp)
        return self


def compute_table_scores(scores: list[Score]) -> float:
    """Compute GW and TPs based on provided VPs, return the maximum VPs scored.
    Check the VPs first with check_table_vps()
    """
    # we're not checking table size here, so try and handle illegal sizes in some way
    # for more than 5 players, just augment the tps 12 by 12
    # table sizes are checked in check_table_vps though, so the score only counts
    # if it gets overriden by a judge
    tps = [t for _, t in zip(range(max(5, len(scores))), itertools.count(12, 12))]
    if len(scores) < 5:
        tps.pop(2)
        # for less than 4, remove the lowest scores, so it stays relevant
        tps = tps[4 - len(scores) :]
    assert len(tps) == len(scores)
    scores_by_vps = sorted(scores, key=lambda x: x.vp)
    # iter over scores in rising order, group by vp count
    for vp, results in itertools.groupby(scores_by_vps, lambda x: x.vp):
        # multiple players can have the same vp count: they shares TPs
        # for the positions they cover
        results = list(results)
        tp = sum(tps.pop(0) for _ in range(len(results))) // len(results)
        # only highest score over 2 (last group = highest vp,
        if vp >= 2 and len(results) == 1 and len(tps) == 0:
            gw = 1
        else:
            gw = 0
        for r in results:
            r.gw = gw
            r.tp = tp
    return vp


class ScoringError(ValueError): ...


class InvalidTableSize(ScoringError): ...


class InsufficientTotal(ScoringError): ...


class ExcessiveTotal(ScoringError): ...


class MissingVP(ScoringError): ...


class MissingHalfVP(ScoringError): ...


def check_table_vps(scores: list[Score]) -> ScoringError | None:
    """Check VPs validity on a table"""
    # check table size
    if len(scores) < 4 or len(scores) > 5:
        return InvalidTableSize()
    # checking the total is easy, just ceil the half points and total is table size
    total = sum(math.ceil(s.vp) for s in scores)
    if total < len(scores):
        return InsufficientTotal()
    if total > len(scores):
        return ExcessiveTotal()
    vps = [[i, s.vp] for i, s in enumerate(scores)]
    # go through all ousts successively: we begin anywhere on the table
    # and search for a zero (which means an oust, otherwise it would be 0.5)
    while len(vps) > 0:
        LOG.debug("scores check pass: %s", vps)
        for j, (idx, vp_count) in enumerate(vps):
            # each oust (vp_count == 0), remove 1 vp from predator ("account" for it)
            # and remove the item from the vps list.
            # Note the predator can have been ousted first (and we're not there yet)
            # so the "accounted" scores can become temporarily negative before
            # being transfered to the next predator
            if vp_count <= 0:
                # vp_count can be negative but should not have half points
                # that would mean we were in a [0.5, <=0] situation previous loop:
                # a half-point score followed by an oust, missing a point
                if vp_count % 1:
                    return MissingVP(f"missing VP for seat {idx + 1}")
                vps[(j - 1) % len(vps)][1] += vp_count - 1
                vps.pop(j)
                break
        # we did not break, all remaining scores are positive
        else:
            # We have some remaining because we checked the total: len(vps) > 0
            # if everyone is at 0.5 it's a timeout
            if all([vp == 0.5 for _, vp in vps]):
                # there must be more than one
                if len(vps) == 1:
                    return MissingHalfVP(f"Seat {vps[0][0] + 1} cannot timeout alone")
            # remove all 0.5
            vps = [[i, vp] for i, vp in vps if vp != 0.5]
            # we can still have one standing if the 0.5 were withdrawals, but not more
            if len(vps) > 1:
                return MissingHalfVP(
                    f"Missing half vps for seats {[i + 1 for i, _ in vps]}"
                )
            # if there is one left, he has 1 point (because of the total check)
            break



import aiohttp
import asyncio
import dataclasses
import datetime
import dotenv
import enum
import logging
import os
import pydantic
import typing
import urllib.parse

from . import engine
from . import geo
from . import models
from . import scoring


dotenv.load_dotenv()
VEKN_LOGIN = os.getenv("VEKN_LOGIN", "")
VEKN_PASSWORD = os.getenv("VEKN_PASSWORD", "")
SITE_URL_BASE = os.getenv("SITE_URL_BASE", "http://127.0.0.1:8000")

LOG = logging.getLogger()


async def get_token(session: aiohttp.ClientSession) -> str:
    async with session.post(
        "https://www.vekn.net/api/vekn/login",
        data={"username": VEKN_LOGIN, "password": VEKN_PASSWORD},
    ) as response:
        try:
            result = await response.json()
            return result["data"]["auth"]
        except KeyError:
            LOG.exception("VEKN authentication failure")
            raise


ADMINS = {
    "3200340",
    "3200188",
    "8180022",
    "3190007",
    "2050001",
    "1002480",
}
JUDGES = {
    "8180022": models.MemberRole.JUDGE,
    "3200188": models.MemberRole.JUDGE,
    "3190007": models.MemberRole.JUDGE,
    "4200005": models.MemberRole.JUDGE,
    "8530107": models.MemberRole.JUDGE,
    "2340000": models.MemberRole.JUDGE,
    "6260014": models.MemberRole.JUDGE,
    "1940030": models.MemberRole.JUDGE,
    "1003731": models.MemberRole.JUDGE,
    "1003455": models.MemberRole.JUDGE,
    "3200340": models.MemberRole.JUDGE,
    "1003030": models.MemberRole.JUDGE,
    "3070069": models.MemberRole.JUDGE,
    "4960027": models.MemberRole.JUDGE,
    "2810001": models.MemberRole.JUDGE,
    "3190133": models.MemberRole.JUDGE,
    "3190041": models.MemberRole.JUDGE,
    "8030009": models.MemberRole.JUDGE,
    "9510021": models.MemberRole.JUDGE,
    "3370036": models.MemberRole.JUDGE,
    "1000629": models.MemberRole.JUDGE,
    "1002855": models.MemberRole.ANC_JUDGE,
    "3340152": models.MemberRole.ANC_JUDGE,
    "5360022": models.MemberRole.JUDGE,
    "8390001": models.MemberRole.NEO_JUDGE,
    "3070006": models.MemberRole.ANC_JUDGE,
    "4960046": models.MemberRole.NEO_JUDGE,
    "6140001": models.MemberRole.NEO_JUDGE,
    "3020044": models.MemberRole.ANC_JUDGE,
    "3020010": models.MemberRole.NEO_JUDGE,
    "1003584": models.MemberRole.ANC_JUDGE,
    "1003214": models.MemberRole.NEO_JUDGE,
    "4110004": models.MemberRole.ANC_JUDGE,
    "4110113": models.MemberRole.ANC_JUDGE,
    "4100033": models.MemberRole.ANC_JUDGE,
    "2331000": models.MemberRole.ANC_JUDGE,
    "3680057": models.MemberRole.ANC_JUDGE,
    "4100008": models.MemberRole.NEO_JUDGE,
    "3120101": models.MemberRole.ANC_JUDGE,
    "4960000": models.MemberRole.NEO_JUDGE,
    "3010501": models.MemberRole.NEO_JUDGE,
    "6060022": models.MemberRole.ANC_JUDGE,
    "5540005": models.MemberRole.NEO_JUDGE,
    "3530067": models.MemberRole.JUDGE,
}


FIX_CITIES = {
    "Argentina": {"BUenos Aires": "Buenos Aires", "Buenos Aries": "Buenos Aires"},
    "Australia": {
        "Blacktown": "Sydney",
        "Castle Hill": "Sydney",
        "Hobart (Rosny)": "Hobart",
        "Hobart, Tasmania": "Hobart",
        "Penrith": "Sydney",
        "Queanbeyan": "Canberra",
        "Ravenhall": "Melbourne",
        "Sydney (Inner City)": "Sydney",
        "Tenambit": "Maitland",
    },
    "Austria": {
        "Danube city (Vienna)": "Vienna",
        "Marchtrenk": "Linz",
        "Thalheim": "Linz",
        "Traiskirchen": "Vienna",
        "Vienna (Traiskirchen)": "Vienna",
        "Wien": "Vienna",
        "Wien/Vienna": "Vienna",
    },
    "Belarus": {
        "Gomel": "Homyel",
    },
    "Belgium": {
        "Antwerp": "Antwerpen",
        "Bruges": "Brugge",
        "Bruxelles": "Brussels",
        "Ghent": "Gent",
        "Lige": "Liège",
        "Liege": "Liège",
    },
    "Brazil": {
        "Brasilia": "Brasília",
        "Braslia": "Brasília",
        "Campinas": "Campinas, São Paulo",
        "Campogrande": "Campina Grande",
        "Canoas / Porto Alegre": "Canoas",
        "GUARULHOS": "Guarulhos",
        "Itajai": "Itajaí",
        "Imperatiz": "Imperatriz",
        "Nova Iguaçú": "Nova Iguaçu",
        "Olaria": "Rio de Janeiro",
        "Petropolis": "Petrópolis",
        "Rio De Janerio": "Rio de Janeiro",
        "Rio de Janerio": "Rio de Janeiro",
        "Rio de janeiro": "Rio de Janeiro",
        "Santo Andre": "Santo André",
        "Sao Bernardo do Campo": "São Bernardo do Campo",
        "São Luis": "São Luís",
        "São PAulo": "São Paulo",
        "So Paulo": "São Paulo",
        "Taguatinga": "Brasília",
        "Vitória / Vila Velha / Grande Vitória": "Vitória",
        "Vitria": "Vitória",
    },
    "Canada": {
        "Edmaonton": "Edmonton",
        "Edmonton / St. Albert": "St. Albert",
        "Edmonton / Spruce Grove": "Spruce Grove",
        "Ednomton": "Edmonton",
        "Gibbons / Edmonton": "Edmonton",
        "Hull": "Gatineau",
        "Jonquiere": "Saguenay",
        "Jonquière": "Saguenay",
        "Levis": "Lévis",
        "Marie Ville": "Montréal",
        "Marieville": "Montréal",
        "Montral": "Montréal",
        "Montreal": "Montréal",
        "Niagara": "Niagara Falls",
        "Qubec City": "Québec",
        "Qubec": "Québec",
        "Quebec": "Québec",
        "Scarborough": "Scarborough Village",
        "St. Albert / Edmonton": "St. Albert",
        "St Catharines": "Sainte-Catherine, Quebec, Montérégie",
        "St Catherines": "Sainte-Catherine, Quebec, Montérégie",
        "St. Catherines": "Sainte-Catherine, Quebec, Montérégie",
        "St-Eustache": "Saint-Eustache",
        "St Eustache": "Saint-Eustache",
        "St. Hubert": "Longueuil",
        "Saint-Hubert": "Longueuil",
        "St-Jean-sur-Richelieu": "Saint-Jean-sur-Richelieu",
        "St-Jerome": "Saint-Jérôme",
        "St-Lazare": "Saint-Lazare",
    },
    "Chile": {
        "Concepcin": "Concepción",
        "Concepcion": "Concepción",
        "Entre Juegos, Santiago": "Santiago",
        "Magic Sur, Santiago": "Santiago",
        "Maip": "Santiago",
        "Quilpue": "Quilpué",
        "Santiago de Chile": "Santiago",
        "Santiago (primogénito)": "Santiago",
        "TableCat Games / Rancagua": "Rancagua",
        "Valparaiso": "Valparaíso",
        "Vina del Mar": "Viña del Mar",
    },
    "Colombia": {"Bogata": "Bogotá", "Bogota": "Bogotá", "Medellin": "Medellín"},
    "Czech Republic": {
        "Brmo": "Brno",
        "Hradec Kralove": "Hradec Králové",
        "Hradec Krlov": "Hradec Králové",
        "Nachod": "Náchod",
        "Plzen": "Pilsen",
        "Praha": "Prague",
        "Slany": "Slaný",
        "Trutnov, Mal Svatoovice": "Trutnov",
        "Vsetin": "Vsetín",
        "Zlin": "Zlín",
    },
    "Denmark": {"Aarhus": "Århus", "Arhus": "Århus"},
    "Finland": {
        "Hyvinkää": "Hyvinge",
        "Kuusankoski": "Kouvola",
    },
    "France": {
        "Alès ": "Alès",
        "Alès / Aix en provence": "Alès",
        "Saint Dizier": "Saint-Dizier",
    },
    "Germany": {
        "Cologne": "Köln",
        "Dsseldorf": "Düsseldorf",
        "Duesseldorf": "Düsseldorf",
        "Frankfurt": "Frankfurt am Main",
        "Gttingen": "Göttingen",
        "Hanau": "Hanau am Main",
        "Ludwigshafen": "Ludwigshafen am Rhein",
        "Madgeburg": "Magdeburg",
        "Marburg": "Marburg an der Lahn",
        "Moerfelden": "Mörfelden-Walldorf",
        "Seeheim": "Seeheim-Jugenheim",
        "Sttutgart": "Stuttgart",
        "Stuttgart / Ludwigsburg": "Ludwigsburg",
    },
    "Greece": {
        "Athens, Attica": "Athens",
        "Athnes": "Athens",
        "Chania": "Chaniá",
        "Thessaloniki": "Thessaloníki",
        "Thessaoniki": "Thessaloníki",
    },
    "Hungary": {
        "debrecen": "Debrecen",
        "Debrechen": "Debrecen",
        "Erdőkertes": "Budapest",
        "Godollo": "Gödöllő",
        "Kaposvar": "Kaposvár",
        "Kecskemet": "Kecskemét",
        "Kismaros": "Budapest",
        "Nyiregyhaza": "Nyíregyháza",
        "Pecs": "Pécs",
        "Salgotarjan": "Salgótarján",
        "Salgtarjn": "Salgótarján",
        "Szekesfehervar": "Székesfehérvár",
        "Szkesfehrvr": "Székesfehérvár",
        "Trnok": "Budapest",
        "Veszprem": "Veszprém",
        "Veszprm": "Veszprém",
    },
    "Iceland": {"Reykjavik": "Reykjavík", "Reykjaví­k": "Reykjavík"},
    "Israel": {"Bat-Yam": "Bat Yam", "Tel-Aviv": "Tel Aviv"},
    "Italy": {
        "Firenze": "Florence",
        "Reggio Emilia": "Reggio nell'Emilia",
        "Torino": "Turin",
        "Milano": "Milan",
        "Genova": "Genoa",
    },
    "Japan": {"Anjo": "Anjō", "Sendai": "Sendai, Miyagi"},
    "Mexico": {
        "Ciudad de México ": "Mexico City",
        "Ciudad de México": "Mexico City",
        "Distrito Federal": "Mexico City",
        "Durango": "Victoria de Durango",
        "Durango, Durango": "Victoria de Durango",
        "Guadalajara, jalisco": "Guadalajara",
        "Monterey, N.L.": "Monterrey",
        "Naucalpan": "Naucalpan de Juárez",
        "Neza": "Ciudad Nezahualcoyotl",
        "Nezahualcoyotl": "Ciudad Nezahualcoyotl",
        "Nezahualcóyotl": "Ciudad Nezahualcoyotl",
        "Puebla": "Puebla, Puebla",
        "Puebla de Zaragoza": "Puebla, Puebla",
        "Queretaro": "Santiago de Querétaro",
        "Toluca de Lerdo": "Toluca",
        "Toluca De Lerdo": "Toluca",
    },
    "Netherlands": {
        "Houten": "Utrecht",
        "Krommenie": "Zaanstad",
        "Rotterdan": "Rotterdam",
    },
    "New Zealand": {"WELLINGTON": "Wellington", "Plamerston North": "Palmerston North"},
    "Norway": {"Fjellhamar": "Oslo"},
    "Panama": {"Panama": "Panamá"},
    "Philippines": {
        "Bacolod": "Bacolod City",
        "Caloocan": "Caloocan City",
        "Dasmarinas, Cavite": "Dasmariñas",
        "Las Pias": "Las Piñas",
        "Los Banos": "Los Baños",
        "Los Baos": "Los Baños",
        "Makati": "Makati City",
        "Marikina": "Marikina City",
        "Metro Manila": "Manila",
        "Parañaque City": "Paranaque City",
        "Quezon": "Quezon City",
        "Quezon city": "Quezon City",
        "Quezon City, Metro Manila": "Quezon City",
        "Taguig City": "Taguig",
        "Tondo, Manila": "Manila",
    },
    "Poland": {
        "Aleksandrow Lodzki": "Aleksandrów Łódzki",
        "Andrespol": "Łódź",
        "Bedzin": "Będzin",
        "Bialystok": "Białystok",
        "Białstok": "Białystok",
        "Bielsko Biaa": "Bielsko-Biala",
        "Bielsko Biała": "Bielsko-Biala",
        "Bielsko-Biała": "Bielsko-Biala",
        "Bielsko-Biaa": "Bielsko-Biala",
        "Boleawiec": "Bolesławiec",
        "Bolesawiec": "Bolesławiec",
        "Cracow": "Kraków",
        "Cracov": "Kraków",
        "Czstochowa": "Częstochowa",
        "Czestochowa": "Częstochowa",
        "Hajnowka": "Hajnówka",
        "Jelenia Gora": "Jelenia Góra",
        "Kędzierzyn Koźle": "Kędzierzyn-Koźle",
        "Krakw": "Kraków",
        "Krakow": "Kraków",
        "Kraszew": "Łódź",
        "Lodz": "Łódź",
        "Lubon": "Luboń",
        "Nowa Sol": "Nowa Sól",
        "Poznan": "Poznań",
        "Swidnik": "Świdnik",
        "Szczezin": "Szczecin",
        "Toru": "Toruń",
        "Torun": "Toruń",
        "Wroclaw": "Wrocław",
    },
    "Portugal": {"Lisboa": "Lisbon", "Setubal": "Setúbal", "Setbal": "Setúbal"},
    "Russian Federation": {
        "Moskow": "Moscow",
        "Saint-Petersburg": "Saint Petersburg",
        "St. Peterburg": "Saint Petersburg",
    },
    "Slovakia": {
        "Banska Bystrica": "Banská Bystrica",
        "Godollo": "Gödöllő",
        "Kosice": "Košice",
    },
    "Spain": {
        "Barberá del Vallés": "Barberà del Vallès",
        "Barcellona": "Barcelona",
        "Barcelona ": "Barcelona",
        "Cádiz": "Cadiz",
        "Castellón de la Plana": "Castelló de la Plana",
        "Castellón": "Castelló de la Plana",
        "Córdoba ": "Córdoba",
        "Gerona": "Girona",
        "Hospitalet de Llobregat": "L'Hospitalet de Llobregat",
        "La Coruña": "A Coruña",
        "Las Palmas": "Las Palmas de Gran Canaria",
        "Las Palmas de Gran Canarias": "Las Palmas de Gran Canaria",
        "Lucena (Córdoba)": "Lucena",
        "Madirid": "Madrid",
        "Masnou": "El Masnou",
        "Mollet del Vallés": "Mollet del Vallès",
        "Palma de Mallorca": "Palma",
        "Rentería": "Errenteria",
        "San Pedro de Alcántara": "Marbella",
        "San Sebastián": "San Sebastián de los Reyes",
        "Sant Cugat del Vallés": "Sant Cugat",
        "Sant Quirze del Vallés": "Sant Quirze del Vallès",
        "Santa Coloma de Gramanet": "Santa Coloma de Gramenet",
        "Sóller": "Palma",
        "Villafranca de Córdoba": "Córdoba",
        "Vitoria": "Gasteiz / Vitoria",
        "Vitoria-Gasteiz": "Gasteiz / Vitoria",
    },
    "Sweden": {
        "Malmo": "Malmö",
        "Örnsköldsviks": "Örnsköldsvik",
        "Stockholm ": "Stockholm",
    },
    "Switzerland": {"Geneva": "Genève", "Zurich": "Zürich"},
    "Ukraine": {"Kiev": "Kyiv"},
    "United States": {
        "ABQ": "Albuquerque",
        "Albuqueruqe": "Albuquerque",
        "Cincinnatti": "Cincinnati",
        "Cinncinati": "Cincinnati",
        "denver": "Denver",
        "Indanapolis": "Indianapolis",
        "Las vegas": "Las Vegas",
        "Los Angelas": "Los Angeles",
        "Los Angleles": "Los Angeles",
        "Mililani": "Mililani Town",
        "New York": "New York City",
        "NYC": "New York City",
        "Palm Bay, FL": "Palm Bay",
        "San Fransisco": "San Francisco",
        "St. George": "Saint George",
        "St Louis": "St. Louis",
        "St. Paul": "Saint Paul",
        "St Paul": "Saint Paul",
        "Saint peters": "Saint Peters",
        "Washington": "Washington, District of Columbia",
        "Washington, D.C.": "Washington, District of Columbia",
    },
    "United Kingdom": {
        "Burton-On-Trent": "Burton upon Trent",
        "Burton-on-Trent": "Burton upon Trent",
        "Burton-on-trent": "Burton upon Trent",
        "Burton-onTrent": "Burton upon Trent",
        "Burton on Trent": "Burton upon Trent",
        "Ealing": "London",
        "Flint, Wales": "Liverpool",
        "Kings Lynn": "King's Lynn",
        "Milton keynes": "Milton Keynes",
        "Newcastle": "Newcastle upon Tyne",
        "Newcastle-Upon-Tyne": "Newcastle upon Tyne",
        "Newcastle Upon-Tyne": "Newcastle upon Tyne",
        "Newcastle Upon Tyne": "Newcastle upon Tyne",
        "Newcastle upon tyne": "Newcastle upon Tyne",
        "Newport, South Wales": "Newport, Wales",
        "Northhampton": "Northampton",
        "Notttingham": "Nottingham",
        "Rochester, Kent": "Rochester",
        "Shefield": "Sheffield",
        "St. Albans": "St Albans",
        "St. Andrews": "Saint Andrews",
        "St. Helens": "St Helens",
        "St. Neots": "Saint Neots",
    },
}


def _member_from_vekn_data(data: dict[str, str]) -> models.Member:
    if data["countryname"]:
        country = geo.COUNTRIES_BY_ISO[data["countrycode"]]
        if data["city"]:
            data_city = data["city"].strip()
            city = geo.CITIES_BY_COUNTRY[country.country].get(data_city, None)
            if not city and data.get("statename", None):
                refined_name = ", ".join([data_city, data["statename"]])
                city = geo.CITIES_BY_COUNTRY[country.country].get(refined_name, None)
            if not city:
                fix = FIX_CITIES.get(country.country, {})
                city = geo.CITIES_BY_COUNTRY[country.country].get(
                    fix.get(data_city, data_city), None
                )
            if not city:
                LOG.info('Did not find city "%s" in %s', data_city, country.country)
        else:
            city = None
    else:
        country = None
        city = None

    roles = []
    prefix = ""
    if data["veknid"] in ADMINS:
        roles.append(models.MemberRole.ADMIN)
    if data.get("coordinatorid", None):
        roles.append(models.MemberRole.NC)
        prefix = data["coordinatorid"]
    if data.get("princeid", None):
        roles.append(models.MemberRole.PRINCE)
        prefix = prefix or data["princeid"]
    judge_role = JUDGES.get(data["veknid"], None)
    if judge_role:
        roles.append(judge_role)
    # TODO maybe keep track of the sponsor/recruit relationship
    # since we're dropping prince and nc prefixes
    return models.Member(
        vekn=data["veknid"],
        name=(data["firstname"] + " " + data["lastname"]).strip(),
        country=country.country if country else "",
        country_flag=country.flag if country else "",
        city=city.unique_name if city else "",
        roles=roles,
        prefix=prefix,
    )


async def get_members_batches() -> typing.AsyncIterator[list[models.Member]]:
    # a few players have a number starting with zero, so start there
    prefix = "00"
    async with aiohttp.ClientSession() as session:
        token = await get_token(session)
        while prefix:
            async with session.get(
                f"https://www.vekn.net/api/vekn/registry?filter={prefix}",
                headers={"Authorization": f"Bearer {token}"},
            ) as response:
                response.raise_for_status()
                result = await response.json()
                players = result["data"]["players"]
                LOG.debug("prefix: %s — %s", prefix, len(players))
                if players:
                    yield [_member_from_vekn_data(data) for data in players]
                # if < 100 players we got them all, just increment the prefix directly
                if len(players) < 100:
                    prefix = increment(prefix)
                # the API returns 100 players max, there might be more
                else:
                    LOG.debug("Last ID: %s", players[-1]["veknid"])
                    prefix = players[-1]["veknid"][:5]
                    if players[-1]["veknid"][-2:] == "99":
                        prefix = increment(prefix)
                # VEKN api will return an empty list on a single-char prefix
                # make sure 59 -> 60 and not 6
                if prefix and len(prefix) < 2:
                    prefix += "0"
                # VEKN api will (wrongly) return an empty list on a "99" prefix
                # because it adds one then pads... careful with the end condition
                if prefix and prefix == "9" * len(prefix) and len(prefix) < 7:
                    prefix += "0"


async def get_events(
    members: list[models.Person],
) -> typing.AsyncIterator[models.Tournament]:
    members = {m.vekn: m for m in members}
    async with aiohttp.ClientSession() as session:
        token = await get_token(session)
        # parallelize by batches of 10
        for num in range(0, 1400):
            tasks = []
            async with asyncio.TaskGroup() as tg:
                for digit in range(0, 10):
                    event_id = 10 * num + digit
                    # skip zero
                    if not event_id:
                        continue
                    tasks.append(
                        tg.create_task(get_event(session, token, event_id, members))
                    )
            for digit, task in enumerate(tasks, 1):
                if not task.done() | task.cancelled():
                    continue
                exc = task.exception()
                if exc:
                    LOG.exception("Failed to retrieve event %s", event_id)
                    continue
                res = task.result()
                if res:
                    yield res


async def get_event(
    session: aiohttp.ClientSession, token: str, num: int, members: list[models.Person]
) -> models.Tournament:
    async with session.get(
        f"https://www.vekn.net/api/vekn/event/{num}",
        headers={"Authorization": f"Bearer {token}"},
    ) as response:
        response.raise_for_status()
        result = await response.json()
        data = result["data"]["events"]
        if not data:
            LOG.info("No data for event #%s: %s", num, result)
            return
        data = data[0]
        if data["players"]:
            LOG.debug("Event #%s: %s", num, data)
            return _tournament_from_vekn_data(data, members)


def _tournament_from_vekn_data(
    data: any, members: dict[str, models.Person]
) -> models.Tournament:
    try:
        fmt, rank = TOURNAMENT_TYPE_TO_FORMAT_RANK[int(data["eventtype_id"])]
    except KeyError:
        LOG.warning(
            "Error in event #%s - unknown event type: %s",
            data["event_id"],
            data["eventtype_id"],
        )
        fmt, rank = models.TournamentFormat.Limited, models.TournamentRank.BASIC
    if data["venue_country"] and data["venue_country"] in geo.COUNTRIES_BY_ISO:
        country = geo.COUNTRIES_BY_ISO[data["venue_country"]].country
    else:
        country = None
    start = " ".join([data["event_startdate"], data["event_starttime"]])
    try:
        start = datetime.datetime.fromisoformat(start)
    except ValueError:
        LOG.info("Error in event #%s - invalid start: %s", data["event_id"], start)
        start = datetime.datetime.fromisoformat(data["event_startdate"])
    finish = " ".join([data["event_enddate"], data["event_endtime"]])
    try:
        finish = datetime.datetime.fromisoformat(finish)
    except ValueError:
        LOG.info("Error in event #%s - invalid finish: %s", data["event_id"], finish)
        finish = datetime.datetime.fromisoformat(data["event_enddate"])
    if data["rounds"]:
        rounds = min(1, int(data["rounds"][0]))
    else:
        rounds = 1
    judges = []
    person = members.get(data["organizer_veknid"])
    if person:
        judges = [models.PublicPerson(**dataclasses.asdict(person))]
    ret = models.Tournament(
        extra={"vekn_id": data["event_id"]},
        name=data["event_name"],
        format=fmt,
        start=start,
        finish=finish,
        timezone="UTC",
        rank=rank,
        country=country,
        venue=data["venue_name"] or "",
        online=bool(int(data["event_isonline"])),
        state=models.TournamentState.FINISHED,
        decklist_required=False,
        proxies=bool(rank == models.TournamentRank.BASIC),
        judges=judges,
    )
    for idx, pdata in enumerate(data["players"], 1):
        member = members.get(pdata["veknid"])
        if not member:
            continue
        try:
            result = scoring.Score(
                gw=int(pdata["gw"]) + (1 if pdata["pos"] == "1" else 0),
                vp=float(pdata["vp"]) + float(pdata["vpf"]),
                tp=int(pdata["tp"]),
            )
        except pydantic.ValidationError:
            LOG.warning(
                "Error in event #%s - invalid player result: %s",
                data["event_id"],
                pdata,
            )
            result = scoring.Score()
        player_rounds = rounds
        # mark 1 round played for DQ, not 0
        # because no-shows are simply not listed in the vekn.net archon
        # DQ or WD only happens if at least a round was played
        # so the player counts in the participants count
        # this matters for the ratings of finalists
        if pdata["dq"] != "0" or pdata["wd"] != "0":
            player_rounds = 1
        elif int(pdata["pos"]) < 6:
            player_rounds += 1
        ret.players[member.uid] = models.Player(
            name=member.name,
            vekn=member.vekn,
            uid=member.uid,
            country=member.country,
            country_flag=member.country_flag,
            city=member.city,
            roles=member.roles,
            sponsor=member.sponsor,
            state=models.PlayerState.FINISHED,
            rounds_played=player_rounds,
            result=result,
            toss=int(pdata["tie"]),
        )
        if pdata["pos"] == "1":
            ret.winner = member.uid
        if pdata["dq"] != "0":
            ret.players[member.uid].barriers.append(models.Barrier.DISQUALIFIED)
    return ret


def increment(num: str) -> str:
    """Increment a numeric prefix. Used to query all members from the VEKN API.

    3012 -> 3013
    3019 -> 302
    ...
    38 -> 39
    39 -> 4
    ...
    98 -> 99
    99 -> None
    """
    while num and num[-1] == "9":
        num = num[:-1]
    if num:
        # keep the leading zeros
        return num[:-1] + str(int(num[-1]) + 1)
    return None


async def get_rankings() -> dict[str, dict[models.RankingCategoy, int]]:
    try:
        async with aiohttp.ClientSession() as session:
            token = await get_token(session)
            async with session.get(
                f"https://www.vekn.net/api/vekn/ranking",
                headers={"Authorization": f"Bearer {token}"},
            ) as response:
                response.raise_for_status()
                result = await response.json()
                ranking = result["data"]["players"][:1000]
    except aiohttp.ClientError:
        LOG.exception("Ranking unavailable")
        ranking = []
    return {
        player["veknid"]: {
            models.RankingCategoy.CONSTRUCTED_ONSITE: int(
                player.pop("rtp_constructed")
            ),
            models.RankingCategoy.LIMITED_ONSITE: int(player.pop("rtp_limited")),
        }
        for player in ranking
    }


class TournamentType(enum.IntEnum):
    DEMO = 1
    STANDARD_CONSTRUCTED = 2
    LIMITED = 3
    MINI_QUALIFIER = 4
    CONTINENTAL_QUALIFIER = 5
    CONTINENTAL_CHAMPIONSHIP = 6
    NATIONAL_QUALIFIER = 7
    NATIONAL_CHAMPIONSHIP = 8
    STORYLINE = 9
    LAUNCH_EVENT = 10
    BUILD_YOUR_OWN_STORYLINE = 11
    UNSANCTIONED_TOURNAMENT = 12
    LIMITED_NATIONAL_CHAMPIONSHIP = 13
    LIMITED_CONTINENTAL_CHAMPIONSHIP = 14
    GRAND_PRIX = 15


TOURNAMENT_TYPE_TO_FORMAT_RANK = {
    TournamentType.DEMO: (models.TournamentFormat.Limited, models.TournamentRank.BASIC),
    TournamentType.STANDARD_CONSTRUCTED: (
        models.TournamentFormat.Standard,
        models.TournamentRank.BASIC,
    ),
    TournamentType.LIMITED: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.MINI_QUALIFIER: (
        models.TournamentFormat.Standard,
        models.TournamentRank.BASIC,
    ),
    TournamentType.CONTINENTAL_QUALIFIER: (
        models.TournamentFormat.Standard,
        models.TournamentRank.GP,
    ),
    TournamentType.CONTINENTAL_CHAMPIONSHIP: (
        models.TournamentFormat.Standard,
        models.TournamentRank.CC,
    ),
    TournamentType.NATIONAL_QUALIFIER: (
        models.TournamentFormat.Standard,
        models.TournamentRank.BASIC,
    ),
    TournamentType.NATIONAL_CHAMPIONSHIP: (
        models.TournamentFormat.Standard,
        models.TournamentRank.NC,
    ),
    TournamentType.STORYLINE: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.LAUNCH_EVENT: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.BUILD_YOUR_OWN_STORYLINE: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.UNSANCTIONED_TOURNAMENT: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.LIMITED_NATIONAL_CHAMPIONSHIP: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.LIMITED_CONTINENTAL_CHAMPIONSHIP: (
        models.TournamentFormat.Limited,
        models.TournamentRank.BASIC,
    ),
    TournamentType.GRAND_PRIX: (
        models.TournamentFormat.Standard,
        models.TournamentRank.GP,
    ),
}


async def upload_tournament(tournament: models.Tournament, rounds: int) -> None:
    type = TournamentType.UNSANCTIONED_TOURNAMENT
    match (tournament.format, tournament.rank):
        case (models.TournamentFormat.Draft, _):
            TournamentType.LIMITED
        case (models.TournamentFormat.Limited, _):
            TournamentType.LIMITED
        case (
            models.TournamentFormat.Standard,
            models.TournamentRank.BASIC,
        ):
            type = TournamentType.STANDARD_CONSTRUCTED
        case (
            models.TournamentFormat.Standard,
            models.TournamentRank.NC,
        ):
            type = TournamentType.NATIONAL_CHAMPIONSHIP
        case (
            models.TournamentFormat.Standard,
            models.TournamentRank.GP,
        ):
            type = TournamentType.CONTINENTAL_QUALIFIER
        case (
            models.TournamentFormat.Standard,
            models.TournamentRank.CC,
        ):
            type = TournamentType.CONTINENTAL_CHAMPIONSHIP
    start = tournament.start.astimezone(tz=datetime.timezone.utc)
    if tournament.finish:
        finish = tournament.finish.astimezone(tz=datetime.timezone.utc)
    else:
        finish = start + datetime.timedelta(hours=6)
    try:
        async with aiohttp.ClientSession() as session:
            token = await get_token(session)
            data = aiohttp.FormData(
                {
                    "name": tournament.name[:120],
                    "type": type,
                    "venueid": 0 if tournament.online else 3800,
                    "online": int(tournament.online),
                    "startdate": start.date().isoformat(),
                    "starttime": f"{start:%H:%M}",
                    "enddate": finish.date().isoformat(),
                    "endtime": f"{finish:%H:%M}",
                    "timelimit": "2h",
                    "rounds": rounds,
                    "final": True,
                    "multideck": tournament.multideck,
                    "proxies": tournament.proxies,
                    "website": urllib.parse.urljoin(
                        SITE_URL_BASE, f"/tournament/{tournament.uid}/display.html"
                    ),
                    "description": tournament.description[:1000],
                }
            )
            event_id = None
            async with session.post(
                "https://www.vekn.net/api/vekn/event",
                headers={"Authorization": f"Bearer {token}"},
                data=data,
            ) as response:
                response.raise_for_status()
                result = await response.json()
                LOG.info("VEKN answered: %s", result)
                if result["data"]["code"] != 200:
                    raise RuntimeError(
                        f"VEKN error: {result['data'].get('message', 'Unknown error')}"
                    )
                event_id = result["data"]["id"]
                tournament.extra["vekn_id"] = event_id
    except aiohttp.ClientError:
        LOG.exception("VEKN Upload failed")
        raise


async def upload_tournament_result(
    tournament: models.Tournament, token: str | None = None
) -> None:
    try:
        async with aiohttp.ClientSession() as session:
            if not token:
                token = await get_token(session)
            async with session.post(
                f"https://www.vekn.net/api/vekn/archon/{tournament.extra['vekn_id']}",
                headers={"Authorization": f"Bearer {token}"},
                data=aiohttp.FormData({"archondata": to_archondata(tournament)}),
            ) as response:
                response.raise_for_status()
                result = await response.json()
                LOG.info("VEKN Archon answered: %s", result)
                if result["data"]["code"] != 200:
                    raise RuntimeError(
                        f"VEKN error: {result['data'].get('message', 'Unknown error')}"
                    )
                tournament.extra["vekn_submitted"] = True
    except aiohttp.ClientError:
        LOG.exception("VEKN Upload failed")
        raise


def to_archondata(tournament: models.Tournament) -> str:
    ret = f"{len(tournament.rounds)}¤"
    if tournament.state != models.TournamentState.FINISHED or not tournament.rounds:
        raise ValueError("Invalid tournament")
    ratings = engine.ratings(tournament)
    finals_table = {s.player_uid: s for s in tournament.rounds[-1].tables[0].seating}
    for rank, player in engine.standings(tournament):
        first, last = (
            player.name.split(" ", 1) if " " in player.name else (player.name, "")
        )
        if player.uid in finals_table:
            final_vp = finals_table[player.uid].result.vp
        else:
            final_vp = 0
        ret += (
            f"{rank}§{first}§{last}§{player.city}§{player.vekn}§{player.result.gw}§{player.result.vp}§{final_vp}"
            f"§{player.result.tp}§{player.toss}§{ratings[player.uid].rating_points}§"
        )
    return ret



import enum
import uuid
import pydantic
import typing
from pydantic import dataclasses


def uuid_str() -> str:
    return str(uuid.uuid4())


class EventType(enum.StrEnum):
    REGISTER = "REGISTER"
    OPEN_CHECKIN = "OPEN_CHECKIN"
    CHECK_IN = "CHECK_IN"
    CHECK_EVERYONE_IN = "CHECK_EVERYONE_IN"
    CHECK_OUT = "CHECK_OUT"
    ROUND_START = "ROUND_START"
    ROUND_ALTER = "ROUND_ALTER"
    ROUND_FINISH = "ROUND_FINISH"
    ROUND_CANCEL = "ROUND_CANCEL"
    SET_RESULT = "SET_RESULT"
    SET_DECK = "SET_DECK"
    DROP = "DROP"
    SANCTION = "SANCTION"
    UNSANCTION = "UNSANCTION"
    OVERRIDE = "OVERRIDE"
    UNOVERRIDE = "UNOVERRIDE"
    SEED_FINALS = "SEED_FINALS"
    SEAT_FINALS = "SEAT_FINALS"
    FINISH_TOURNAMENT = "FINISH_TOURNAMENT"


class SanctionLevel(enum.StrEnum):
    CAUTION = "Caution"
    WARNING = "Warning"
    DISQUALIFICATION = "Disqualification"
    BAN = "Ban"


class SanctionCategory(enum.StrEnum):
    NONE = ""
    DECK_PROBLEM = "Deck Problem"
    PROCEDURAL_ERROR = "Procedural Error"
    CARD_DRAWING = "Card drawing"
    MARKED_CARDS = "Marked Cards"
    SLOW_PLAY = "Slow Play"
    UNSPORTSMANLIKE_CONDUCT = "Unsportsmanlike Conduct"
    CHEATING = "Cheating"
    ETHICS = "Ethics"


@dataclasses.dataclass(kw_only=True)
class Event:
    type: EventType
    uid: str = pydantic.Field(default_factory=uuid_str)


@dataclasses.dataclass(kw_only=True)
class Register(Event):
    type: typing.Literal[EventType.REGISTER]
    name: str
    vekn: str = ""
    player_uid: str = pydantic.Field(default_factory=uuid_str)
    country: str | None = ""
    city: str | None = ""


@dataclasses.dataclass(kw_only=True)
class OpenCheckin(Event):
    type: typing.Literal[EventType.OPEN_CHECKIN]


@dataclasses.dataclass(kw_only=True)
class CheckIn(Event):
    type: typing.Literal[EventType.CHECK_IN]
    player_uid: str
    code: str = ""


@dataclasses.dataclass(kw_only=True)
class CheckEveryoneIn(Event):
    type: typing.Literal[EventType.CHECK_EVERYONE_IN]


@dataclasses.dataclass(kw_only=True)
class CheckOut(Event):
    type: typing.Literal[EventType.CHECK_OUT]
    player_uid: str


@dataclasses.dataclass(kw_only=True)
class RoundStart(Event):
    type: typing.Literal[EventType.ROUND_START]
    seating: list[list[str]]  # list of tables: [[player_uid]]


@dataclasses.dataclass(kw_only=True)
class RoundAlter(Event):
    type: typing.Literal[EventType.ROUND_ALTER]
    round: int
    seating: list[list[str]]  # list of tables: [[player_uid]]


@dataclasses.dataclass(kw_only=True)
class RoundFinish(Event):
    type: typing.Literal[EventType.ROUND_FINISH]


@dataclasses.dataclass(kw_only=True)
class RoundCancel(Event):
    type: typing.Literal[EventType.ROUND_CANCEL]


@dataclasses.dataclass(kw_only=True)
class SetResult(Event):
    type: typing.Literal[EventType.SET_RESULT]
    player_uid: str
    round: int
    vps: float


@dataclasses.dataclass(kw_only=True)
class SetDeck(Event):
    type: typing.Literal[EventType.SET_DECK]
    player_uid: str
    round: int | None = None  # Can have a different deck each round if multideck
    deck: str  # Deck builder URL or plain text decklist


@dataclasses.dataclass(kw_only=True)
class Drop(Event):
    type: typing.Literal[EventType.DROP]
    player_uid: str


@dataclasses.dataclass(kw_only=True)
class Sanction(Event):
    type: typing.Literal[EventType.SANCTION]
    level: SanctionLevel
    sanction_uid: str
    player_uid: str
    comment: str = ""
    category: SanctionCategory | None = None


@dataclasses.dataclass(kw_only=True)
class Unsanction(Event):
    type: typing.Literal[EventType.UNSANCTION]
    player_uid: str
    sanction_uid: str


@dataclasses.dataclass(kw_only=True)
class Override(Event):
    type: typing.Literal[EventType.OVERRIDE]
    round: int
    table: int
    comment: str


@dataclasses.dataclass(kw_only=True)
class Unoverride(Event):
    type: typing.Literal[EventType.UNOVERRIDE]
    round: int
    table: int


@dataclasses.dataclass(kw_only=True)
class SeedFinals(Event):
    type: typing.Literal[EventType.SEED_FINALS]
    toss: dict[str, int]  # {player_uid; toss}
    seeds: list[str]  # [player_uid] in seed order (first is top seed)


@dataclasses.dataclass(kw_only=True)
class SeatFinals(Event):
    type: typing.Literal[EventType.SEAT_FINALS]
    seating: list[str]  # [player_uid]


@dataclasses.dataclass(kw_only=True)
class FinishTournament(Event):
    type: typing.Literal[EventType.FINISH_TOURNAMENT]


TournamentEvent = typing.Union[
    Register,
    OpenCheckin,
    CheckIn,
    CheckEveryoneIn,
    CheckOut,
    RoundStart,
    RoundAlter,
    RoundFinish,
    RoundCancel,
    SetResult,
    SetDeck,
    Drop,
    Sanction,
    Unsanction,
    Override,
    Unoverride,
    SeedFinals,
    SeatFinals,
    FinishTournament,
]


OPENAPI_EXAMPLES = {
    "Register": {
        "summary": "Register a player",
        "description": (
            "Neither VEKN nor UID is mandatory. "
            "To register a new player who has no VEKN account, provide a new UUID4. "
            "If you do not provide one, "
            "a new UUID4 will be generated and an account created for that person."
        ),
        "value": {
            "type": "Register",
            "name": "John Doe",
            "vekn": "12300001",
            "player_uid": "24AAC87E-DE63-46DF-9784-AB06B2F37A24",
        },
    },
    "Open Check-in": {
        "summary": "Open the check-in. JUDGE ONLY",
        "description": (
            "Check a player in, signaling they are present "
            "and ready to play the next round."
            "You should perform the check-in just before the round starts to limit "
            "the number of players who do not show up to their table."
        ),
        "value": {"type": "OpenCheckin"},
    },
    "Appoint Judge": {
        "summary": "Appoint a judge. JUDGE ONLY",
        "description": (
            "Judges have permission to send all events, simple players are limited. "
            "The UID must match a VEKN member UID."
        ),
        "value": {
            "type": "AppointJudge",
            "judge_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
        },
    },
    "Remove Judge": {
        "summary": "Remove a judge. JUDGE ONLY",
        "description": (
            "Judges have permission to send all events, simple players are limited. "
            "The UID must match a VEKN member UID."
        ),
        "value": {
            "type": "RemoveJudge",
            "judge_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
        },
    },
    "Check In": {
        "summary": "Check a player in",
        "description": "Mark a player as ready to play. Players can self-check-in. ",
        "value": {
            "type": "CheckIn",
            "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
        },
    },
    "Check Everyone In": {
        "summary": "Check everyone in. JUDGE ONLY",
        "description": (
            "When running registrations in situ, or after first round. "
            "It will not check-in players who have dropped (FINISHED state) "
            "or have an active barrier (missing deck, having been disqualified, etc.)"
        ),
        "value": {
            "type": "CheckEveryoneIn",
        },
    },
    "Round Start": {
        "summary": "Starts the next round. JUDGE ONLY",
        "description": (
            "Only possible if the previous round is finished. "
            "The provided seating must list players UID forming the tables. "
            "Each UID must match a VEKN member UID. "
        ),
        "value": {
            "type": "RoundStart",
            "seating": [
                [
                    "238CD960-7E54-4A38-A676-8288A5700FC8",
                    "796CD3CE-BC2B-4505-B448-1C2D42E9F140",
                    "80E9FD37-AD8C-40AA-A42D-138065530F10",
                    "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
                    "8F28E4C2-1953-473E-A1C5-C281957072D1",
                ],
                [
                    "BD570AA9-B70C-43CA-AD05-3B4C7DADC28C",
                    "AB6F75B3-ED60-45CA-BDFF-1BF8DD5F02C4",
                    "1CB1E9A7-576B-4065-8A9C-F7920AAF977D",
                    "8907BE41-91A7-4395-AF91-54D94C489A36",
                ],
            ],
        },
    },
    "Round Alter": {
        "summary": "Changes a round seating. JUDGE ONLY",
        "description": (
            "Rounds are numbered starting with 1. You can alter a finished round."
            "The provided seating must list players UID forming the tables. "
            "Each UID must match a VEKN member UID. "
        ),
        "value": {
            "type": "RoundAlter",
            "round": 1,
            "seating": [
                [
                    "238CD960-7E54-4A38-A676-8288A5700FC8",
                    "796CD3CE-BC2B-4505-B448-1C2D42E9F140",
                    "80E9FD37-AD8C-40AA-A42D-138065530F10",
                    "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
                    "8F28E4C2-1953-473E-A1C5-C281957072D1",
                ],
                [
                    "BD570AA9-B70C-43CA-AD05-3B4C7DADC28C",
                    "AB6F75B3-ED60-45CA-BDFF-1BF8DD5F02C4",
                    "1CB1E9A7-576B-4065-8A9C-F7920AAF977D",
                    "8907BE41-91A7-4395-AF91-54D94C489A36",
                ],
            ],
        },
    },
    "Round Finish": {
        "summary": "Finish the current round. JUDGE ONLY",
        "description": (
            "Once a round is finished you cannot re-open it, but you can still modify "
            "it with RoundAlter, SetResult and Override events"
        ),
        "value": {"type": "RoundFinish"},
    },
    "Round Cancel": {
        "summary": "Cancel the current round. JUDGE ONLY",
        "description": (
            "A round can be canceled, if no result has been set yet. "
            "Beware this removes the current seating and cannot be undone."
            "Can be used to cancel a final's seeding, but will not reset the toss data."
        ),
        "value": {"type": "RoundCancel"},
    },
    "Set Result": {
        "summary": "Set a player's result",
        "description": (
            "Players can set their and their table result for the current round. "
            "Only VPs are provided, the GW and TP computations are done by the engine."
        ),
        "value": {
            "type": "SetResult",
            "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
            "round": 1,
            "vps": 2.5,
        },
    },
    "Set Deck": {
        "summary": "Set a player's deck",
        "description": (
            "Players can set their deck for the tournament, "
            "or the next round for multideck tournaments. "
            "The `deck` field can contain a plaintext deck list, or a deckbuilder URL "
            "(VDB, Amaranth, VTESDecks)"
        ),
        "value": {
            "type": "SetDeck",
            "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
            "deck": "https://vdb.im/decks/11808",
        },
    },
    "Drop": {
        "summary": "Drop a player from the tournament.",
        "description": (
            "A player can drop by themselves. "
            "A Judge can drop a player if they note they have juse left. "
            "To **disqualify** a player, use the Sanction event."
        ),
        "value": {"type": "Drop", "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8"},
    },
    "Sanction": {
        "summary": "Sanction (punish) a player. JUDGE ONLY",
        "description": (
            "The sanction levels are: `CAUTION`, `WARNING` and `DISQUALIFICATION`."
            "Cautions are just informative. "
            "Warnings are recorded (accessible to organizers, even in future events). "
            "Disqualifications are recorded and remove the player from the tournament. "
            "Sanction also have an optional category, one of: \n"
            "- ``\n"
            "- `DECK_PROBLEM`\n"
            "- `PROCEDURAL_ERRORS`\n"
            "- `CARD_DRAWING`\n"
            "- `MARKED_CARDS`\n"
            "- `SLOW_PLAY`\n"
            "- `UNSPORTSMANLIKE_CONDUCT`\n"
            "- `CHEATING`\n"
        ),
        "value": {
            "type": "Sanction",
            "level": "WARNING",
            "sanction_uid": "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
            "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
            "comment": "Free comment",
            "category": "PROCEDURAL_ERRORS",
        },
    },
    "Unsanction": {
        "summary": "Remove all sanctions of given level for a player. JUDGE ONLY",
        "description": (
            "Judges have permission to send all events, simple players are limited. "
            "The player UID must match a VEKN member UID. "
            "The sanction UID must match an existing sanction UID for that player. "
        ),
        "value": {
            "type": "Unsanction",
            "sanction_uid": "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
            "player_uid": "238CD960-7E54-4A38-A676-8288A5700FC8",
        },
    },
    "Override": {
        "summary": "Override a table score. JUDGE ONLY",
        "description": (
            "Judges can validated an odd table score. For example, if they disqualify "
            "a player but do not award VPs to their predator, the final table score "
            "will not appear valid until it's overriden.\n"
            "Rounds and tables are counted starting from 1."
        ),
        "value": {
            "type": "Override",
            "round": 1,
            "table": 1,
            "comment": "Free form comment",
        },
    },
    "Seed Finals": {
        "summary": "Seed the finals. JUDGE ONLY",
        "description": (
            "A finals is seeded first before players elect their seat in seed order."
        ),
        "value": {
            "type": "SeedFinals",
            "seeds": [
                "238CD960-7E54-4A38-A676-8288A5700FC8",
                "796CD3CE-BC2B-4505-B448-1C2D42E9F140",
                "80E9FD37-AD8C-40AA-A42D-138065530F10",
                "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
                "8F28E4C2-1953-473E-A1C5-C281957072D1",
            ],
        },
    },
    "Seat Finals": {
        "summary": "Seat the finals. JUDGE ONLY",
        "description": ("Note what seating position finalists have elected"),
        "value": {
            "type": "SeatFinals",
            "seating": [
                "238CD960-7E54-4A38-A676-8288A5700FC8",
                "796CD3CE-BC2B-4505-B448-1C2D42E9F140",
                "80E9FD37-AD8C-40AA-A42D-138065530F10",
                "586616DC-3FEA-4DAF-A222-1E77A2CBD809",
                "8F28E4C2-1953-473E-A1C5-C281957072D1",
            ],
        },
    },
    "Finish": {
        "summary": "Finish the tournament. JUDGE ONLY",
        "description": (
            "This closes up the tournament. "
            "The winner, if finals results have been recorded, "
            "is automatically computed."
        ),
        "value": {
            "type": "Finish",
        },
    },
}



import dataclasses
import fastapi
import logging
import typing

from .. import dependencies
from ... import events
from ... import models

LOG = logging.getLogger()
router = fastapi.APIRouter(
    prefix="/api/tournaments",
    default_response_class=fastapi.responses.ORJSONResponse,
    tags=["tournaments"],
)


@router.get("/", summary="List all tournaments")
async def api_tournaments(
    filter: typing.Annotated[models.TournamentFilter, fastapi.Query()],
    op: dependencies.DbOperator,
) -> tuple[models.TournamentFilter, list[models.TournamentMinimal]]:
    """List all tournaments"""
    if filter.uid or filter.country or not filter.online or filter.states:
        filter = filter
    else:
        filter = None
    return await op.get_tournaments_minimal(filter)


@router.post("/", summary="Create a new tournament")
async def api_tournaments_post(
    request: fastapi.Request,
    data: typing.Annotated[models.TournamentConfig, fastapi.Body()],
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
) -> dependencies.TournamentUrl:
    """Create a new tournament"""
    dependencies.check_organizer(member)
    data.judges = await op.get_members(
        list(set([j.uid for j in data.judges]) | {member.uid})
    )
    LOG.info("Creating new tournament: %s", data)
    uid = await op.create_tournament(data)
    return dependencies.TournamentUrl(
        uid=uid, url=str(request.url_for("tournament_display", uid=uid))
    )


@router.put("/{uid}", summary="Update tournament information")
async def api_tournament_put(
    request: fastapi.Request,
    orchestrator: dependencies.TournamentOrchestrator,
    data: typing.Annotated[models.TournamentConfig, fastapi.Body()],
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
) -> dependencies.TournamentUrl:
    """Update tournament information

    - **uid**: The tournament unique ID
    """
    data.judges = await op.get_members([judge.uid for judge in data.judges])
    orchestrator.update_config(data, member)
    uid = await op.update_tournament(orchestrator)
    return dependencies.TournamentUrl(
        uid=uid, url=str(request.url_for("tournament_display", uid=uid))
    )


@router.get("/{uid}", summary="Get tournament data")
async def api_tournament_get(
    tournament: dependencies.Tournament, member: dependencies.PersonFromToken
) -> models.Tournament:
    """Get tournament data

    - **uid**: The tournament unique ID
    """
    dependencies.check_can_admin_tournament(member, tournament)
    return tournament


@router.get("/{uid}/info", summary="Get tournament public information")
async def api_tournament_get_info(
    tournament: dependencies.Tournament, member: dependencies.PersonFromToken
) -> models.Tournament | models.TournamentInfo:
    """Get tournament information

    - **uid**: The tournament unique ID
    """
    return models.TournamentInfo(**dataclasses.asdict(tournament))


@router.get("/{uid}/decks", summary="Get tournament decks information")
async def api_tournament_get(
    tournament: dependencies.Tournament, _: dependencies.MemberUidFromToken
) -> models.TournamentDeckInfo:
    """Get tournament information

    - **uid**: The tournament unique ID
    """
    res = models.TournamentDeckInfo(**dataclasses.asdict(tournament))
    if tournament.multideck:
        for round_ in tournament.rounds:
            for table in round_.tables:
                for seat in table.seating:
                    if seat.deck:
                        res.decks.append(seat.deck)
    else:
        for player in tournament.players.values():
            if player.deck:
                res.decks.append(player.deck)
    return res


@router.post(
    "/{uid}/vekn-sync/{rounds}", summary="Sync finished tournament to vekn.net"
)
async def vekn_sync(
    tournament: dependencies.Tournament,
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    rounds: typing.Annotated[int, fastapi.Path()],
) -> models.Tournament:
    dependencies.check_can_admin_tournament(member, tournament)
    await dependencies.vekn_sync(tournament, rounds)
    await op.update_tournament(tournament)
    return tournament


@router.post("/{uid}/set-vekn/{vekn_id}", summary="Create tournament on vekn.net")
async def set_vekn(
    tournament: dependencies.Tournament,
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    vekn_id: typing.Annotated[str, fastapi.Path()],
) -> models.Tournament:
    dependencies.check_can_admin_tournament(member, tournament)
    tournament.extra["vekn_id"] = vekn_id
    await op.update_tournament(tournament)
    return tournament


@router.delete("/{uid}", summary="Delete tournament")
async def api_tournament_delete(
    tournament: dependencies.Tournament,
    actor: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
) -> None:
    """Delete tournament

    - **uid**: The tournament unique ID
    """
    if models.MemberRole.ADMIN not in actor.roles:
        raise fastapi.HTTPException(fastapi.status.HTTP_403_FORBIDDEN)
    for player_uid in tournament.players:
        member: models.Member = await op.get_member(player_uid, True)
        for i, sanction in enumerate(member.sanctions):
            if sanction.tournament and sanction.tournament.uid == tournament.uid:
                member.sanctions.pop(i)
        member.ratings.pop(tournament.uid, None)
        await op.update_member(member)
    await op.delete_tournament(tournament.uid)


@router.post("/{uid}/event", summary="Add tournament event")
async def api_tournament_event_post(
    orchestrator: dependencies.TournamentOrchestrator,
    event: typing.Annotated[
        events.TournamentEvent, fastapi.Body(openapi_examples=events.OPENAPI_EXAMPLES)
    ],
    actor: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
) -> models.Tournament:
    """Send a new event for this tournament.

    This is the main way of interacting with a tournament data.

    - **uid**: The tournament unique ID
    """
    orchestrator.handle_event(event, actor)
    await op.record_event(orchestrator.uid, actor.uid, event)
    await op.update_tournament(orchestrator)
    # TODO: we might want to move this in a "VEKN member orchestrator" of sorts
    if (
        event.type == events.EventType.SANCTION
        and event.level != events.SanctionLevel.CAUTION
    ):
        member = await op.get_member(event.player_uid, for_update=True)
        member.sanctions.append(
            models.RegisteredSanction(
                tournament=models.TournamentConfig(**dataclasses.asdict(orchestrator)),
                uid=event.sanction_uid,
                judge=actor,
                level=event.level,
                category=event.category,
                comment=event.comment,
            )
        )
        await op.update_member(member)
    if event.type == events.EventType.UNSANCTION:
        member = await op.get_member(event.player_uid, for_update=True)
        for idx, sanction in enumerate(member.sanctions):
            if sanction.uid == event.sanction_uid:
                del member.sanctions[idx]
        await op.update_member(member)
    if event.type == events.EventType.FINISH_TOURNAMENT:
        await dependencies.vekn_sync(orchestrator, max(1, len(orchestrator.rounds)))
        await op.update_tournament(orchestrator)
    return orchestrator



import dataclasses
import fastapi
import typing

from .. import dependencies
from ... import geo
from ... import models


router = fastapi.APIRouter(
    prefix="/api/vekn",
    default_response_class=fastapi.responses.ORJSONResponse,
    tags=["vekn"],
)


@router.get("/country", summary="List all countries")
async def api_vekn_countries() -> list[models.Country]:
    """List all countries"""
    return sorted(geo.COUNTRIES_BY_NAME.values(), key=lambda c: c.country)


@router.get("/country/{country}/city", summary="List cities of given country")
async def api_vekn_country_cities(
    country: typing.Annotated[str, fastapi.Path()]
) -> list[models.City]:
    """List cities of given country.

    Only cities **over 15k population** are listed.
    Open-source information made availaible by [Geonames](https://geonames.org).

    - **country**: The country name, field `country` of `/api/countries`
    """
    if country not in geo.CITIES_BY_COUNTRY:
        raise fastapi.HTTPException(fastapi.status.HTTP_404_NOT_FOUND)
    return sorted(geo.CITIES_BY_COUNTRY[country].values(), key=lambda c: c.unique_name)


@router.post("/claim")
async def api_vekn_claim(
    param: typing.Annotated[models.VeknParameter, fastapi.Body()],
    member_uid: dependencies.MemberUidFromToken,
    op: dependencies.DbOperator,
) -> dependencies.Token:
    new_member = await op.claim_vekn(member_uid, param.vekn)
    if new_member is None:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_403_FORBIDDEN, detail="VEKN unknown or claimed already"
        )
    access_token = dependencies.create_access_token(new_member.uid)
    return dependencies.Token(access_token=access_token, token_type="Bearer")


@router.post("/abandon")
async def api_vekn_abandon(
    member_uid: dependencies.MemberUidFromToken, op: dependencies.DbOperator
) -> dependencies.Token:
    new_member = await op.abandon_vekn(member_uid)
    if new_member is None:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_404_NOT_FOUND, detail="No VEKN token"
        )
    access_token = dependencies.create_access_token(new_member.uid)
    return dependencies.Token(access_token=access_token, token_type="Bearer")


@router.get("/members")
async def api_vekn_members(
    member: dependencies.PersonFromToken, op: dependencies.DbOperator
) -> list[models.Person]:
    if member.vekn:
        return await op.get_members()
    else:
        return await op.get_externally_visible_members(member)


@router.get("/members/{uid}")
async def api_vekn_member(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
) -> models.Member:
    ret = await op.get_member(uid)
    if member.uid != uid and not member.vekn:
        dependencies.check_can_contact(member, ret)
    if not ret:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_404_NOT_FOUND, detail="No VEKN token"
        )
    return ret


@router.post("/members")
async def api_vekn_add_member(
    posting_member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    member: typing.Annotated[models.Member, fastapi.Body()],
) -> models.Member:
    dependencies.check_organizer(posting_member)
    if member.country:
        if member.country in geo.COUNTRIES_BY_NAME:
            country = geo.COUNTRIES_BY_NAME[member.country]
            member.country = country.country
            member.country_flag = country.flag
        else:
            member.country = ""
            member.country_flag = ""
    city = None
    if member.country and member.city:
        city = geo.CITIES_BY_COUNTRY[member.country].get(member.city, None)
    if city:
        member.city = city.unique_name
    else:
        member.city = ""
    return await op.insert_member(member)


@router.post("/members/password")
async def api_vekn_set_member_password(
    member: dependencies.MemberFromToken,
    op: dependencies.DbOperator,
    password: typing.Annotated[models.PasswordParameter, fastapi.Body()],
) -> models.Member:
    dependencies.set_member_password(member, password.password)
    return await op.update_member(member)


@router.post("/members/unlink_discord")
async def api_vekn_unlink_discord(
    member: dependencies.MemberFromToken,
    op: dependencies.DbOperator,
) -> models.Member:
    member.discord = None
    return await op.update_member(member)


@router.post("/members/{uid}/add_role")
async def api_vekn_member_add_role(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    param: typing.Annotated[models.RoleParameter, fastapi.Body()],
) -> models.Member:
    target = await op.get_member(uid, True)
    dependencies.check_can_change_role(member, target, param.role)
    target.roles = list(set(target.roles) | {param.role})
    return await op.update_member(target)


@router.post("/members/{uid}/remove_role")
async def api_vekn_member_remove_role(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    param: typing.Annotated[models.RoleParameter, fastapi.Body()],
) -> models.Member:
    target = await op.get_member(uid, True)
    dependencies.check_can_change_role(member, target, param.role)
    target.roles = list(set(target.roles) - {param.role})
    return await op.update_member(target)


@router.post("/members/{uid}/sponsor")
async def api_vekn_member_sponsor(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
) -> models.Member:
    dependencies.check_organizer(member)
    target = await op.get_member(uid, True)
    target.sponsor = member.uid
    return await op.update_member_new_vekn(target)


@router.post("/members/{uid}/vekn")
async def api_vekn_member_assign_vekn(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    param: typing.Annotated[models.VeknParameter, fastapi.Body()],
) -> models.Member:
    target = await op.get_member(uid, True)
    dependencies.check_can_change_vekn(member, target)
    target.sponsor = member.uid
    return await op.claim_vekn(target.uid, param.vekn)


@router.delete("/members/{uid}/vekn")
async def api_vekn_member_delete_vekn(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
) -> models.Member:
    target = await op.get_member(uid, True)
    dependencies.check_can_change_vekn(member, target)
    return await op.abandon_vekn(target.uid)


@router.post("/members/{uid}/info")
async def api_vekn_member_info(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    info: typing.Annotated[models.MemberInfo, fastapi.Body()],
) -> models.Member:
    target = await op.get_member(uid, True)
    dependencies.check_can_change_info(member, target)
    for field in dataclasses.fields(info):
        value = getattr(info, field.name)
        if value is None:
            continue
        setattr(target, field.name, value)
    return await op.update_member(target)


@router.post("/members/{uid}/sanction")
async def api_vekn_member_sanction(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    sanction: typing.Annotated[models.RegisteredSanction, fastapi.Body()],
) -> models.Member:
    dependencies.check_can_sanction(member)
    sanction.judge_uid = member.uid
    target = await op.get_member(uid, True)
    if sanction.tournament:
        # for now, tournament sanctions should be delivered through tournament events
        raise fastapi.HTTPException(fastapi.status.HTTP_400_BAD_REQUEST)
    target.sanctions.append(sanction)
    return await op.update_member(target)


@router.delete("/members/{uid}/sanction/{sanction_uid}")
async def api_vekn_member_sanction_delete(
    member: dependencies.PersonFromToken,
    op: dependencies.DbOperator,
    uid: typing.Annotated[str, fastapi.Path()],
    sanction_uid: typing.Annotated[str, fastapi.Path()],
) -> models.Member:
    dependencies.check_can_sanction(member)
    target = await op.get_member(uid, True)
    if not any(s.uid == sanction_uid for s in target.sanctions):
        raise fastapi.HTTPException(fastapi.status.HTTP_404_NOT_FOUND)
    target.sanctions = [s for s in target.sanctions if s.uid != sanction_uid]
    return await op.update_member(target)
